# -*- coding: UTF-8 -*-

import sciv

if __name__ == '__main__':

    # The save path for process files and result files.
    save_path = "/path/result/data"
    # The save path for PoissonVI training model.
    poisson_vi_save_path = "/path/result/data/poisson_vi"
    cache_path = "/path/result/cache_path"
    log_path = "/path/result/log"

    # Path containing three files
    sc_atac_base_path = "/path/data/meta"
    # A folder for storing trait or disease files
    variant_base_path = "/path/data/variant"

    # set log information
    sciv.ul.is_form_log_file = True
    sciv.ul.log_file_path = log_path

    # set cache path
    sciv.ul.project_cache_path = cache_path

    # read variant information
    variants, trait_info = sciv.fl.read_variants(base_path=variant_base_path)

    # scATAC-seq data
    sc_atac = sciv.fl.read_sc_atac(resource=sc_atac_base_path)

    # run
    trs = sciv.ml.core(
        adata=sc_atac,
        variants=variants,
        trait_info=trait_info,
        model_dir=poisson_vi_save_path,
        save_path=save_path,
        is_file_exist_loading=True
    )

    print(trs)



# -*- coding: UTF-8 -*-

import os
from typing import Union, Tuple, Literal
from pathlib import Path

import matplotlib
from numpy import ndarray, matrix
from pandas import CategoricalDtype
from matplotlib import rcParams, colormaps
from matplotlib.colors import ListedColormap
from PyComplexHeatmap import colors

from scipy.sparse import (
    coo_matrix,
    csr_matrix,
    csc_matrix,
    dok_matrix,
    lil_matrix,
    bsr_matrix,
    dia_matrix,
    spmatrix,
    coo_array,
    csr_array,
    csc_array,
    dok_array,
    lil_array,
    bsr_array,
    dia_array,
    sparray
)

project_version = "0.0.1"
project_name = "SCIV"

user_path = os.path.expanduser("~")
project_cache_path: str = os.path.join(user_path, ".cache", project_name)

is_form_log_file: bool = False
log_file_path: str = os.path.join(project_cache_path, "log")

matplotlib.set_loglevel("error")

path = Union[str, Path]
number = Union[int, float]
collection = Union[list, set, Tuple, ndarray]

sparse_array = Union[coo_array, csr_array, csc_array, dok_array, lil_array, bsr_array, dia_array, sparray]
sparse_matrix = Union[coo_matrix, csr_matrix, csc_matrix, dok_matrix, lil_matrix, bsr_matrix, dia_matrix, spmatrix]

sparse_data = Union[sparse_array, sparse_matrix]
dense_data = Union[ndarray, matrix, list]

matrix_data = Union[sparse_data, dense_data]

enrichment_optional = Literal['golden', 'half', 'e', 'pi', 'none']
difference_peak_optional = Literal['emp_effect', 'bayes_factor', 'emp_prob1', 'all']

plot_rc_config = {
    "font.family": 'Arial',
    "axes.labelsize": 7,
    "font.size": 7,
    "legend.fontsize": 7,
    "pdf.fonttype": 42,
    "ps.fonttype": 42,
    "axes.unicode_minus": False
}
rcParams.update(plot_rc_config)

colors.define_cmap()
type_50_colors = colormaps.get("cmap50").colors.copy()
plot_cmap_50 = ListedColormap(type_50_colors)

type_20_colors = list(colormaps.get("tab20").colors).copy()
plot_cmap_20 = ListedColormap(type_20_colors)

type_set_colors = [
    "#66c2a5", "#fc8d62", "#8da0cb", "#e78ac3", "#a6d854", "#f4d037",
    "#e5c494", "#f2f0a7", "#9a60b4", "#c16772", "#c1c167", "#b3b3b3",
    "#4b0082", "#cd853f", "#008080", "#808000", "#ff00ff", "#2f4f4f",
    "#d3b48b", "#8b4513", "#7fff00", "#ff69b4", "#66cdaa", "#a52a2a",
    "#daa520"
]
plot_cmap_set = ListedColormap(type_set_colors)

plot_color_types: dict = {
    "20": type_20_colors,
    "50": type_50_colors,
    "set": type_set_colors
}

chrtype = CategoricalDtype(
    [
        "chr1",
        "chr2",
        "chr3",
        "chr4",
        "chr5",
        "chr6",
        "chr7",
        "chr8",
        "chr9",
        "chr10",
        "chr11",
        "chr12",
        "chr13",
        "chr14",
        "chr15",
        "chr16",
        "chr17",
        "chr18",
        "chr19",
        "chr20",
        "chr21",
        "chr22",
        "chrX",
        "chrY"
    ],
    ordered=True
)



# -*- coding: UTF-8 -*-

import os
import time
import warnings
from typing import Optional

import numpy as np
import pandas as pd

from anndata import AnnData
from torch.cuda import OutOfMemoryError

from .. import util as ul
from ..tool import umap, tsne
from ..util import path, check_gpu_availability

__name__: str = "preprocessing_scvi"


def poisson_vi(
    adata: AnnData,
    max_epochs: int = 500,
    lr: float = 1e-4,
    batch_size: int = 128,
    eps: float = 1e-08,
    early_stopping: bool = True,
    early_stopping_patience: int = 50,
    strategy: str = "ddp_notebook_find_unused_parameters_true",
    batch_key: Optional[str] = None,
    resolution: float = 0.5,
    dp_delta: float = 0.05,
    latent_name: str = "latent",
    model_dir: Optional[path] = None
) -> AnnData:
    """
    PoissonVI processing of the data results in the current sample representation and peak difference data after Leiden clustering.
    :param adata: processing data;
    :param max_epochs: The maximum number of epochs for PoissonVI training;
    :param lr: Learning rate for optimization;
    :param batch_size: Minibatch size to use during training;
    :param eps: Optimizer eps;
    :param early_stopping: Whether to perform early stopping with respect to the validation set;
    :param early_stopping_patience: How many epochs to wait for improvement before early stopping;
    :param strategy: DDP strategy;
    :param batch_key: Batch information in scATAC-seq data;
    :param resolution: Resolution of the Leiden Cluster;
    :param dp_delta: PeakVI method in differential analysis empirical effect size threshold;
    :param latent_name: The name of latent representation;
    :param model_dir: The folder name saved by the training module;
    :return: Differential peak of clustering types.
    """
    ul.log(__name__).info("Start PoissonVI")

    start_time = time.perf_counter()

    import scvi
    import scanpy as sc

    if resolution <= 0:
        ul.log(__name__).error("The parameter `resolution` must be greater than zero.")
        raise ValueError("The parameter `resolution` must be greater than zero.")

    if dp_delta <= 0:
        ul.log(__name__).error("The parameter `dp_delta` must be greater than zero.")
        raise ValueError("The parameter `dp_delta` must be greater than zero.")

    if batch_key is not None and batch_key not in adata.obs.columns:
        ul.log(__name__).error(f"The cells information {adata.obs.columns} in data `adata` must include the {batch_key} column.")
        raise ValueError(f"The cells information {adata.obs.columns} in data `adata` must include the {batch_key} column.")

    # PoissonVI, Binarization
    ul.log(__name__).info("Calculate fragment counts matrix.")
    scvi.data.reads_to_fragments(adata)

    def __train__():

        # PoissonVI
        scvi.external.POISSONVI.setup_anndata(adata, layer="fragments", batch_key=batch_key)
        _model_ = scvi.external.POISSONVI(adata)

        if check_gpu_availability():

            try:
                data_splitter_kwargs = {"drop_dataset_tail": True, "drop_last": False}
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    _model_.train(
                        max_epochs=int(max_epochs),
                        check_val_every_n_epoch=1,
                        accelerator="gpu",
                        devices=-1,
                        datasplitter_kwargs=data_splitter_kwargs,
                        strategy=strategy,
                        lr=lr,
                        batch_size=int(batch_size),
                        eps=eps,
                        early_stopping=early_stopping,
                        early_stopping_patience=int(early_stopping_patience)
                    )
            except Exception as ex:

                try:
                    ul.log(__name__).warning(f"Multiple GPU failed to run, attempting to run on one card.\n {ex}")
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        _model_.train(
                            max_epochs=int(max_epochs),
                            check_val_every_n_epoch=1,
                            lr=lr,
                            batch_size=int(batch_size),
                            eps=eps,
                            early_stopping=early_stopping,
                            early_stopping_patience=int(early_stopping_patience)
                        )
                except Exception as exc:
                    ul.log(__name__).warning(f"GPU failed to run, try to switch to CPU running.\n {exc}")
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        _model_.to_device('cpu')
                        _model_.train(
                            max_epochs=int(max_epochs),
                            check_val_every_n_epoch=1,
                            lr=lr,
                            batch_size=int(batch_size),
                            eps=eps,
                            early_stopping=early_stopping,
                            early_stopping_patience=int(early_stopping_patience),
                            accelerator="cpu"
                        )
        else:

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                _model_.to_device('cpu')
                _model_.train(
                    max_epochs=int(max_epochs),
                    check_val_every_n_epoch=1,
                    lr=lr,
                    batch_size=int(batch_size),
                    eps=eps,
                    early_stopping=early_stopping,
                    early_stopping_patience=int(early_stopping_patience),
                    accelerator="cpu"
                )

        return _model_

    if model_dir is not None:
        if os.path.exists(os.path.join(model_dir, "model.pt")):
            ul.log(__name__).info(f"Due to the existence of file `model.pt`, it is loaded by default.")

            try:
                model = scvi.external.POISSONVI.load(model_dir, adata=adata)
            except OutOfMemoryError as ome:
                ul.log(__name__).warning(f"GPU failed to run, try to switch to CPU running.\n {ome}")

                try:
                    model = scvi.external.POISSONVI.load(model_dir, adata=adata, accelerator="cpu")
                except Exception as e:
                    ul.log(__name__).error(f"File `model.pt` failed to load, you can execute `Poisson VI` again by deleting file `model.pt` ({model_dir}/model.pt).\n {e}")
                    raise ValueError(f"File `model.pt` failed to load, you can execute `Poisson VI` again by deleting file `model.pt` ({model_dir}/model.pt).")
        else:
            ul.file_method(__name__).makedirs(model_dir)
            model = __train__()
            model.save(model_dir, overwrite=True)
    else:
        model = __train__()

    # latent space
    latent = model.get_latent_representation()
    adata.obsm[latent_name] = latent

    if "clusters" in adata.obs.columns:
        ul.log(__name__).warning("Due to the original inclusion of the `clusters` column, the original `clusters` column name has been changed to `clusters_x`.")
        adata.obs["clusters_x"] = adata.obs["clusters"]

    ul.log(__name__).info(f"Perform kNN and Leiden clustering.")
    # compute the k-nearest-neighbor graph that is used in both clustering and umap algorithms
    sc.pp.neighbors(adata, use_rep=latent_name)
    # cluster the space (we use a lower resolution to get fewer clusters than the default)
    sc.tl.leiden(adata, key_added="clusters", resolution=resolution)
    adata.obs["clusters"] = adata.obs["clusters"].astype(str)

    # umap
    try:
        data_umap = umap(adata.obsm[latent_name])
        adata.obsm["umap"] = data_umap
        adata.obs["latent_umap1"] = data_umap[:, 0]
        adata.obs["latent_umap2"] = data_umap[:, 1]
    except Exception as e:
        ul.log(__name__).warning(f"UMAP error, your system does not support it, but it does not affect the process. Continue with execution: {e}")

    # tsne
    try:
        data_tsne = tsne(adata.obsm[latent_name])
        adata.obsm["tsne"] = data_tsne
        adata.obs["latent_tsne1"] = data_tsne[:, 0]
        adata.obs["latent_tsne2"] = data_tsne[:, 1]
    except Exception as e:
        ul.log(__name__).warning(f"TSNE error, your system does not support it, but it does not affect the process. Continue with execution: {e}")

    clusters_list = list(set(adata.obs["clusters"]))
    clusters_list.sort()

    adata.uns["poisson_vi"] = {
        "model_dir": model_dir,
        "cluster_size": len(clusters_list),
        "dp_delta": dp_delta,
        "latent_name": latent_name
    }
    peaks_info = adata.var.copy()
    peaks_info["index"] = peaks_info.index
    peaks_info.rename_axis("peak_index", inplace=True)

    matrix_bf = np.ones((len(clusters_list), adata.shape[1]))
    matrix_ep1 = np.ones((len(clusters_list), adata.shape[1]))
    matrix_ee = np.ones((len(clusters_list), adata.shape[1]))

    if len(clusters_list) == 1:
        ul.log(__name__).warning(f"The number of clusters is one, ignoring cluster type correction. It is recommended to increase the `resolution` value to make it effective.")
    else:
        # differential peak
        da_peaks_all: dict = {}

        for cluster in clusters_list:
            ul.log(__name__).info(f"Start difference peak: {cluster}/({', '.join(clusters_list)})")
            clusters_all = clusters_list.copy()
            clusters_all.remove(cluster)

            # differential peak
            try:
                da_peaks = model.differential_accessibility(adata, groupby="clusters", delta=dp_delta, group1=cluster, mode="vanilla", two_sided=False)
            except Exception as e:
                ul.log(__name__).warning(f"GPU failed to run, try to switch to CPU running.\n {e}")
                # PyTorch uses CPU
                model.to_device('cpu')
                da_peaks = model.differential_accessibility(adata, groupby="clusters", delta=dp_delta, group1=cluster, mode="vanilla", two_sided=False)

            da_peaks_all.update({cluster: da_peaks})

        adata.uns["da_peaks"] = da_peaks_all

        for i in range(len(clusters_list)):
            cluster_info = da_peaks_all[clusters_list[i]]
            cluster_info["index"] = cluster_info.index
            cluster_info.rename_axis("cluster_index", inplace=True)
            cluster_info = pd.merge(left=peaks_info, right=cluster_info, left_on="index", right_on="index", how="left")
            matrix_bf[i, :] = cluster_info["bayes_factor"]
            matrix_ep1[i, :] = cluster_info["emp_prob1"]
            matrix_ee[i, :] = cluster_info["emp_effect"]

    ul.log(__name__).info("End PoissonVI")

    obs = pd.DataFrame(clusters_list, columns=["id"])
    obs.index = obs["id"].astype(str)

    da_peaks_adata = AnnData(matrix_ee, obs=obs, var=adata.var)
    da_peaks_adata.layers["bayes_factor"] = matrix_bf
    da_peaks_adata.layers["emp_prob1"] = matrix_ep1
    da_peaks_adata.uns["latent_name"] = latent_name
    da_peaks_adata.uns["dp_delta"] = dp_delta
    da_peaks_adata.uns["elapsed_time"] = time.perf_counter() - start_time

    adata.uns["step"] = 1

    return da_peaks_adata



# -*- coding: UTF-8 -*-

import math
import time
from typing import Union, Tuple, Literal

import torch
import torch.nn as nn

from torch import Tensor
from tqdm import tqdm
from joblib import Parallel, delayed

import numpy as np
from anndata import AnnData
from pandas import DataFrame

from . import mean_symmetric_scale, min_max_norm, perturb_data

from .. import util as ul
from ..util import (
    matrix_data,
    to_sparse,
    to_dense,
    collection,
    check_adata_get,
    enrichment_optional,
    check_gpu_availability,
    sparse_data
)

__name__: str = "tool_random_walk"


def _random_walk_cpu_(
    seed_cell_vector: Union[list, np.ndarray, np.matrix],
    weight: matrix_data = None,
    gamma: float = 0.05,
    epsilon: float = 1e-5,
    p: int = 2
) -> collection:
    """
    Perform a random walk
    :param seed_cell_vector: seed cells;
    :param weight: weight matrix;
    :param gamma: reset weight.
    :param epsilon: conditions for stopping in random walk;
    :param p: Distance used for loss {1: Manhattan distance, 2: Euclidean distance};
    :return: The value after random walk.
    """

    # Random walk
    p0 = np.asarray(seed_cell_vector, dtype=float).ravel()[:, np.newaxis]
    pt: matrix_data = p0.copy()
    k = 0
    delta = 1

    # iteration
    while delta > epsilon:

        if hasattr(weight, "dot"):
            p1 = (1 - gamma) * weight.dot(pt) + gamma * p0
        else:
            p1 = (1 - gamma) * np.dot(weight, pt) + gamma * p0

        # 1 and 2, It would be faster alone
        if p == 1:
            delta = np.abs(pt - p1).sum()
        elif p == 2:
            delta = np.sqrt(np.square(np.abs(pt - p1)).sum())
        else:
            delta = np.float_power(np.float_power(np.abs(pt - p1), p).sum(), 1.0 / p)

        pt = p1
        k += 1

    return pt.flatten()


class RandomWalkModel(nn.Module):

    def __init__(self, gamma: float = 0.05, epsilon: float = 1e-5, p: int = 2, device: str = 'auto', pbar=None):
        super().__init__()
        self.gamma = gamma
        self.epsilon = epsilon
        self.p = p
        self.pbar = pbar

        is_gpu_available = check_gpu_availability()

        self.device = 'cuda' if (device == 'gpu' or (device == 'auto' and is_gpu_available)) else 'cpu'

        self.factor = 1 - self.gamma

    def core(self, seed_cell_vector: Tensor, weight: Tensor):

        p0 = seed_cell_vector
        pt = p0.clone()

        delta = 1.0
        k = 0

        while delta > self.epsilon:
            p1 = self.factor * torch.matmul(weight, pt) + self.gamma * p0
            delta = torch.linalg.norm(pt - p1, ord=self.p).item()
            pt = p1
            k += 1

        if self.pbar is not None:
            self.pbar.update(1)

        return pt.flatten()

    def forward(self, seed_cell_weight: Tensor, weight: Tensor):

        sample_count = seed_cell_weight.shape[1]

        score = torch.zeros(seed_cell_weight.shape).cuda()

        for i in range(sample_count):
            score[:, i] = self.core(seed_cell_weight[:, i], weight)

        return score


class TraitDataParallel(nn.DataParallel):

    def scatter(self, inputs, kwargs, device_ids):
        _seed_cell_weight_, _weight_ = inputs
        scattered_seed_cell_weight = torch.nn.parallel.scatter(_seed_cell_weight_, device_ids, dim=1)
        scattered_weight = [_weight_.to(f'cuda:{device_id}') for device_id in device_ids]
        scattered_inputs = [(ssw, sw) for ssw, sw in zip(scattered_seed_cell_weight, scattered_weight)]

        scattered_kwargs = []

        for device_id in device_ids:
            device_kwargs = {}

            for key, value in kwargs.items():

                if isinstance(value, torch.Tensor):
                    device_kwargs[key] = value.to(f'cuda:{device_id}')
                else:
                    device_kwargs[key] = value

            scattered_kwargs.append(device_kwargs)

        return scattered_inputs, scattered_kwargs

    def gather(self, outputs, output_device):
        """
        Collect the results after parallel processing, check for the existence of results,
        and merge the results by column (each result matrix has the same number of rows but different numbers of columns)
        :param outputs: Output results of each device
        :param output_device: output device
        :return: Collect and merge the results by column
        """
        return torch.nn.parallel.scatter_gather.gather(outputs, output_device, dim=1)


def _random_walk_gpu_(
    seed_cell_weight: matrix_data,
    weight: matrix_data,
    gamma: float = 0.05,
    epsilon: float = 1e-5,
    p: int = 2,
    device: str = 'auto'
) -> matrix_data:

    with tqdm(total=seed_cell_weight.shape[1]) as pbar:

        model = RandomWalkModel(gamma, epsilon, p, device, pbar)

        device = model.device
        model.to(device)

        seed_cell_weight = torch.as_tensor(seed_cell_weight, device=device, dtype=torch.float32)
        weight = to_dense(weight, is_array=True)
        weight = torch.as_tensor(weight, device=device, dtype=torch.float32)

        if device == 'cuda' and 1 < torch.cuda.device_count() < seed_cell_weight.shape[1]:
            model = TraitDataParallel(model)

        with torch.no_grad():
            result = model(seed_cell_weight, weight)

        return result.cpu().numpy()


def random_walk(
    seed_cell_weight: matrix_data,
    weight: matrix_data,
    gamma: float = 0.05,
    epsilon: float = 1e-5,
    p: int = 2,
    n_jobs: int = -1,
    device: str = 'auto'
) -> matrix_data:

    availability = check_gpu_availability()

    if device == 'cpu' or (device == 'auto' and not availability):
        sample_count = seed_cell_weight.shape[1]

        results = Parallel(n_jobs=n_jobs)(
            delayed(_random_walk_cpu_)(seed_cell_weight[:, i], weight, gamma, epsilon, p)
            for i in tqdm(range(sample_count))
        )

        return np.column_stack(results)
    elif device == 'gpu' or (device == 'auto' and availability):

        try:
            return _random_walk_gpu_(seed_cell_weight, weight, gamma, epsilon, p, device='gpu')
        except RuntimeError as e:
            ul.log(__name__).warning(f"GPU failed to run, try to switch to CPU running.\n {e}")
            sample_count = seed_cell_weight.shape[1]

            results = Parallel(n_jobs=n_jobs)(
                delayed(_random_walk_cpu_)(seed_cell_weight[:, i], weight, gamma, epsilon, p)
                for i in tqdm(range(sample_count))
            )

            return np.column_stack(results)
    else:
        ul.log(__name__).error(
            f'The `device` ({device}) is not supported. Only supports "cpu", "gpu", and "auto" values.'
        )
        raise ValueError(f'The `device` ({device}) is not supported. Only supports "cpu", "gpu", and "auto" values.')


def trs_scale_norm(score: matrix_data, axis: Literal[0, 1, -1] = 0, is_verbose: bool = True) -> matrix_data:
    cell_value = mean_symmetric_scale(score, axis=axis, is_verbose=is_verbose)
    cell_value = np.log1p(min_max_norm(cell_value, axis=axis))
    return cell_value


class RandomWalk:
    """
    Random walk
    """

    def __init__(
        self,
        cc_adata: AnnData,
        init_status: AnnData,
        epsilon: float = 1e-05,
        gamma: float = 0.05,
        enrichment_gamma: float = 0.05,
        p: int = 2,
        n_jobs: int = -1,
        min_seed_cell_rate: float = 0.01,
        max_seed_cell_rate: float = 0.05,
        credible_threshold: float = 0,
        enrichment_threshold: Union[enrichment_optional, float] = 'golden',
        benchmark_count: int = 10,
        is_ablation: bool = False,
        is_simple: bool = True
    ):
        """
        Perform random walk steps
        :param cc_adata: Cell features;
        :param init_status: For cell scores under each trait;
        :param epsilon: conditions for stopping in random walk;
        :param gamma: reset weight for random walk;
        :param enrichment_gamma: reset weight for random walk for enrichment;
        :param p: Distance used for loss {1: Manhattan distance, 2: Euclidean distance};
        :param n_jobs: The maximum number of concurrently running jobs;
        :param min_seed_cell_rate: The minimum percentage of seed cells in all cells;
        :param max_seed_cell_rate: The maximum percentage of seed cells in all cells.
        :param credible_threshold: The threshold for determining the credibility of enriched cells in the context of
            enrichment, i.e. the threshold for judging enriched cells;
        :param enrichment_threshold: Only by setting a threshold for the standardized output TRS can a portion of the
            enrichment results be obtained. Parameters support string types {'golden', 'half', 'e', 'pi', 'none'}, or
            valid floating-point types within the range of (0, log1p(1)).
        :param is_ablation: True represents obtaining the results of the ablation experiment. This parameter is limited
            by the `is_simple` parameter, and its effectiveness requires setting `is_simple` to `False`;
        :param is_simple: True represents not adding unnecessary intermediate variables, only adding the final result.
            It is worth noting that when set to `True`, the `is_ablation` parameter will become invalid, and when set
            to `False`, `is_ablation` will only take effect;
        :return: Stable distribution score.
        """
        ul.log(__name__).info("Random walk with weighted seed cells.")

        start_time = time.perf_counter()

        # judge length
        if cc_adata.shape[0] != init_status.shape[0]:
            ul.log(__name__).error(
                f"The number of rows {cc_adata.shape[0]} in the data is not equal to the initialization state length "
                f"{np.array(init_status).size}"
            )
            raise ValueError(
                f"The number of rows {cc_adata.shape[0]} in the data is not equal to the initialization state length "
                f"{np.array(init_status).size}"
            )

        if p <= 0:
            ul.log(__name__).error(
                "The value of `p` must be greater than zero. Distance used for loss {1: Manhattan distance, "
                "2: Euclidean distance}"
            )
            raise ValueError(
                "The value of `p` must be greater than zero. Distance used for loss {1: Manhattan distance, "
                "2: Euclidean distance}"
            )
        elif p > 3:
            ul.log(__name__).warning("Suggested value for `p` is 1 or 2.")

        if epsilon > 0.1:
            ul.log(__name__).warning(
                f"Excessive value of parameter `epsilon`=({epsilon}) can lead to incorrect iteration and poor "
                f"enrichment effect."
            )
        elif epsilon <= 0:
            ul.log(__name__).error("The parameter of `epsilon` must be greater than zero.")
            raise ValueError("The parameter of `epsilon` must be greater than zero.")

        if "clusters" not in init_status.obs.columns:
            ul.log(__name__).error(
                "Unsupervised clustering information must be included in column `clusters` of `init_datus.obs`."
            )
            raise ValueError(
                "Unsupervised clustering information must be included in column `clusters` of `init_datus.obs`."
            )

        init_status.obs["clusters"] = init_status.obs["clusters"].astype(str)

        self.epsilon = epsilon
        self.gamma = gamma
        self.enrichment_gamma = enrichment_gamma
        self.p = p
        self.n_jobs = n_jobs
        self.min_seed_cell_rate = min_seed_cell_rate
        self.max_seed_cell_rate = max_seed_cell_rate
        self.credible_threshold = credible_threshold
        self.is_simple = is_simple
        self.is_ablation = is_ablation
        self.benchmark_count = benchmark_count
        self._enrichment_seed_cell_min_count_ = 3

        self.is_gpu_available = check_gpu_availability()

        if not is_simple and self.is_ablation:
            if "cell_mutual_knn" not in cc_adata.layers:
                ul.log(__name__).error("The ablation requires `cell_mutual_knn` to be in `cc_adata.layers`.")
                raise ValueError("The ablation requires `cell_mutual_knn` to be in `cc_adata.layers`.")

        if isinstance(enrichment_threshold, float):

            if enrichment_threshold <= 0 or enrichment_threshold >= np.log1p(1):
                ul.log(__name__).warning(
                    "The `enrichment_threshold` parameter is not set within the range of (0, log1p(1)), "
                    "this parameter will become invalid."
                )
                ul.log(__name__).warning(
                    "It is recommended to set the `enrichment_threshold` parameter to the 'golden' value."
                )

            self.enrichment_threshold = enrichment_threshold
        elif enrichment_threshold == "golden":
            golden_ratio = (1 + math.sqrt(5)) / 2
            self.enrichment_threshold = np.log1p(1) / (1 + 1 / golden_ratio)
        elif enrichment_threshold == "half":
            self.enrichment_threshold = np.log1p(1) / 2
        elif enrichment_threshold == "e":
            self.enrichment_threshold = np.log1p(1) / np.e
        elif enrichment_threshold == "pi":
            self.enrichment_threshold = np.log1p(1) / np.pi
        elif enrichment_threshold == "none":
            self.enrichment_threshold = np.log1p(1)
        else:
            raise ValueError(
                "Invalid enrichment settings. The string type in the `enrichment_threshold` parameter only supports "
                "the following parameter 'golden', 'half', 'e', 'pi',  Alternatively, input a floating-point type "
                "value within the range of (0, log1p(1))"
            )

        # Enrichment judgment
        self.is_run_core = False
        self.is_run_ablation_m_knn = False
        self.is_run_ablation_ncw = False
        self.is_run_ablation_nsw = False
        self.is_run_ablation_ncsw = False

        self.is_run_enrichment = False
        self.is_run_en_ablation_m_knn = False
        self.is_run_en_ablation_ncw = False
        self.is_run_en_ablation_nsw = False
        self.is_run_en_ablation_ncsw = False

        self.is_benchmark = False
        self.cluster_size_factor = {}

        self.cell_affinity = to_dense(cc_adata.layers["cell_affinity"])

        self.init_status: AnnData = init_status
        self.trait_info: list = list(init_status.var["id"])

        self.trs_adata: AnnData = AnnData(np.zeros(init_status.shape), obs=init_status.obs, var=init_status.var)
        self.trs_adata.uns = init_status.uns
        self.trs_adata.layers["init_trs"] = to_sparse(init_status.X)

        self.cell_anno = self.trs_adata.obs

        if not is_simple:
            for _layer_ in init_status.layers:
                self.trs_adata.layers[_layer_] = to_sparse(init_status.layers[_layer_])

        self.cell_size: int = self.trs_adata.shape[0]

        # trait
        self.trait_list: list = list(self.trs_adata.var_names)
        self.trait_range = range(len(self.trait_list))

        self.trs_source = np.zeros(init_status.shape)
        self.trs_source_positive = np.zeros(init_status.shape)
        self.trs_source_negative = np.zeros(init_status.shape)

        if not is_simple and self.is_ablation:
            self.trs_m_knn_source = np.zeros(init_status.shape)
            self.trs_ncw_source = np.zeros(init_status.shape)
            self.trs_nsw_source = np.zeros(init_status.shape)
            self.trs_ncsw_source = np.zeros(init_status.shape)
            self.random_seed_cell = np.zeros(init_status.shape)

        # Transition Probability Matrix
        self.weight = self._get_weight_(cc_adata.X)

        if not is_simple and self.is_ablation:
            self.weight_m_knn = self._get_weight_(cc_adata.layers["cell_mutual_knn"])

        del cc_adata

        self.cluster_types, self.init_seed_cell_size = self._get_cluster_info_()

        (
            self.seed_cell_count,
            self.seed_cell_threshold,
            self.seed_cell_weight_nsw,
            self.seed_cell_weight,
            self.seed_cell_index,
            self.seed_cell_weight_en_nsw,
            self.seed_cell_weight_en
        ) = self._get_seed_cell_()

        if not is_simple and self.is_ablation:
            init_status_no_weight = check_adata_get(init_status, "init_trs_ncw")
            (
                self.seed_cell_count_nw,
                self.seed_cell_threshold_nw,
                self.seed_cell_weight_ncsw,
                self.seed_cell_weight_ncw,
                _,
                self.seed_cell_weight_en_ncsw,
                self.seed_cell_weight_en_ncw
            ) = self._get_seed_cell_(init_data=init_status_no_weight, info="ablation")

        del self.cell_affinity
        del init_status

        self.elapsed_time = time.perf_counter() - start_time

    def _random_walk_(
        self,
        seed_cell_data: matrix_data,
        weight: matrix_data = None,
        gamma: float = 0.05,
        device: str = 'auto'
    ) -> matrix_data:
        """
        Perform a random walk
        :param seed_cell_data: seed cells;
        :param weight: weight matrix;
        :param gamma: reset weight.
        :param device: device.
        :return: The value after random walk.
        """

        if weight is None:
            w = self.weight
        else:
            w = weight

        if not self.is_gpu_available:
            return random_walk(
                seed_cell_data, weight=w, gamma=gamma, epsilon=self.epsilon, p=self.p, n_jobs=self.n_jobs, device='cpu'
            )

        try:
            _data_ = random_walk(
                seed_cell_data, weight=w, gamma=gamma, epsilon=self.epsilon, p=self.p, n_jobs=self.n_jobs, device=device
            )
        except Exception as e:
            ul.log(__name__).warning(f"GPU failed to run, try to switch to CPU running.\n {e}")
            _data_ = random_walk(
                seed_cell_data, weight=w, gamma=gamma, epsilon=self.epsilon, p=self.p, n_jobs=self.n_jobs, device='cpu'
            )

        return _data_

    def _random_walk_core_(self, seed_cell_data: matrix_data, weight: matrix_data = None) -> matrix_data:
        """
        Perform a random walk
        :param seed_cell_data: seed cells;
        :param weight: weight matrix.
        :return: The value after random walk.
        """
        return self._random_walk_(seed_cell_data, weight, self.gamma)

    @staticmethod
    def _get_weight_(cell_cell_matrix: matrix_data) -> sparse_data:
        """
        Obtain weights in random walk
        :param cell_cell_matrix: Cell to cell connectivity matrix
        :return: weight matrix
            1. The weights used in the iteration of random walk.
            2. Assign different weight matrices to seed cells.
        """
        ul.log(__name__).info("Obtain transition probability matrix.")
        data_weight = to_dense(cell_cell_matrix, is_array=True)
        cell_sum_weight = data_weight.sum(axis=1)[:, np.newaxis]
        cell_sum_weight[cell_sum_weight == 0] = 1
        return to_sparse(data_weight / cell_sum_weight)

    def _get_cell_weight_(self, seed_cell_size: int) -> matrix_data:
        _cell_cell_knn_: matrix_data = self.cell_affinity.copy()

        # Obtain numerical values for constructing a k-neighbor network
        cell_cell_affinity_sort = np.sort(_cell_cell_knn_, axis=1)
        cell_cell_value = cell_cell_affinity_sort[:, -(seed_cell_size + 1)]
        del cell_cell_affinity_sort
        _cell_cell_knn_[self.cell_affinity < np.array(cell_cell_value).flatten()[:, np.newaxis]] = 0
        return _cell_cell_knn_

    def _get_seed_cell_size_(self, cell_size: int) -> int:
        seed_cell_size: int = self.init_seed_cell_size if self.init_seed_cell_size < cell_size else cell_size

        # Control the number of seeds
        if (seed_cell_size / self.cell_size) < self.min_seed_cell_rate:
            seed_cell_size = np.ceil(self.min_seed_cell_rate * self.cell_size).astype(int)
        elif (seed_cell_size / self.cell_size) > self.max_seed_cell_rate:
            seed_cell_size = np.ceil(self.max_seed_cell_rate * self.cell_size).astype(int)

        if seed_cell_size == 0:
            seed_cell_size = 3
        elif seed_cell_size > cell_size:
            seed_cell_size = cell_size

        return seed_cell_size

    def _get_cluster_info_(self) -> Tuple[list, int]:
        # cluster size/count
        cluster_types = list(set(self.trs_adata.obs["clusters"]))
        cluster_types.sort()

        clusters = list(self.trs_adata.obs["clusters"])

        for cluster in cluster_types:
            count = clusters.count(cluster)
            self.cluster_size_factor.update({str(cluster): count})

        seed_cell_size = min(self.cluster_size_factor.values())

        self.trs_adata.uns["cluster_info"] = {
            "cluster_size_factor": self.cluster_size_factor,
            "min_seed_cell_rate": self.min_seed_cell_rate,
            "max_seed_cell_rate": self.max_seed_cell_rate,
            "init_seed_cell_size": seed_cell_size
        }
        return cluster_types, seed_cell_size

    def _get_seed_cell_clustering_weight_(self, seed_cell_index: collection) -> Tuple[collection, dict]:
        """
        This function is used to obtain the percentage of seed cells that occupy this cell type, i.e., the seed cell
        clustering weight. The purpose of this weight is to provide fair enrichment opportunities for those with fewer
        cell numbers in cell clustering types.
        :param seed_cell_index: Index of seed cells.
        :return: The seed cell clustering weight, equity factor.
        """
        cell_anno: DataFrame = self.cell_anno.copy()
        cell_clusters = cell_anno["clusters"].values
        seed_cell_cell_anno: DataFrame = cell_anno.iloc[seed_cell_index]

        seed_cell_cell_clusters: list = cell_clusters[seed_cell_index].tolist()
        seed_cell_cell_cluster_rate = {}

        for _k_, _v_ in self.cluster_size_factor.items():
            seed_cell_cell_cluster_rate.update({_k_: seed_cell_cell_clusters.count(_k_) / _v_})

        seed_cell_cluster_weight = seed_cell_cell_anno["clusters"].map(seed_cell_cell_cluster_rate).values
        return mean_symmetric_scale(seed_cell_cluster_weight, is_verbose=False), seed_cell_cell_cluster_rate

    def _get_seed_cell_weight_(
        self,
        seed_cell_index: collection,
        value: collection,
        seed_cell_index_enrichment: collection = None
    ) -> collection:

        if seed_cell_index_enrichment is None:
            seed_cell_index_enrichment = seed_cell_index

        # Calculate the degree of seed cells in the seed cell network
        seed_cell_mutual_knn = np.array(self.cell_affinity[seed_cell_index, :][:, seed_cell_index])
        seed_weight_degree: collection = seed_cell_mutual_knn.sum(axis=0)
        seed_weight_degree_weight = mean_symmetric_scale(seed_weight_degree, is_verbose=False)

        # Calculate the initialization score weight
        seed_cell_value = value[seed_cell_index_enrichment]
        seed_cell_value_weight = mean_symmetric_scale(seed_cell_value, is_verbose=False)

        # Percentage weight of seed cells in cell type clustering
        seed_cell_clustering_weight = self._get_seed_cell_clustering_weight_(seed_cell_index)[0]
        seed_weight_value = seed_weight_degree_weight * seed_cell_value_weight * seed_cell_clustering_weight

        # Calculate weight
        seed_weight_value = seed_weight_value / (1 if np.sum(seed_weight_value) == 0 else np.sum(seed_weight_value))
        return seed_weight_value

    def _get_seed_cell_(
        self,
        init_data: AnnData = None,
        info: str = None
    ) -> Tuple[collection, collection, matrix_data, matrix_data, matrix_data, matrix_data, matrix_data]:
        """
        Obtain information related to seed cells
        :param init_data: Initial TRS data
        :param info: Log information about seed cells
        :return:
            1. Set seed cell thresholds for each trait or disease.
            2. Seed cell weights obtained for each trait or disease based on the `init_data` parameter, with each seed
                cell assigned the same weight. Note that this only takes effect when `is_simple` is true.
            3. Seed cell weights obtained for each trait or disease based on the init_data parameter, and the weight of
                each seed cell will be assigned based on the similarity between cells.
            4. Seed cell index, which will be used for later knockout or knockdown prediction.
            5. Based on the init_data parameter, a reference seed cell weight is obtained for enrichment analysis
                assistance for each trait or disease, and each seed cell is assigned the same weight.
                Note that this only takes effect when `is_simple` is true.
            6. Reference seed cell weights for auxiliary enrichment analysis of each trait or disease based on the
                init_data parameter, and the weight of each seed cell will be assigned based on the similarity
                between cells.
        """

        if init_data is None:
            init_data = self.init_status

        n_traits = len(self.trait_list)
        n_cells = self.cell_size

        seed_cell_count = np.zeros(n_traits, dtype=int)
        seed_cell_threshold = np.zeros(n_traits)
        seed_cell_weight = np.zeros((n_cells, n_traits))
        seed_cell_index = np.zeros((n_cells, n_traits), dtype=int)
        seed_cell_weight_en = np.zeros((n_cells, n_traits))

        if not self.is_simple:
            seed_cell_matrix = np.zeros((n_cells, n_traits))
            seed_cell_matrix_en = np.zeros((n_cells, n_traits))
        else:
            seed_cell_matrix = np.zeros((1, 1))
            seed_cell_matrix_en = np.zeros((1, 1))

        ul.log(__name__).info(
            f"Calculate {n_traits} traits/diseases for seed cells information.{f' ({info})' if info else ''}"
        )

        trait_values_all = to_dense(init_data.X, is_array=True)

        def _process_single_trait(i: int) -> None:
            trait_value = trait_values_all[:, i]
            trait_value_max = trait_value.max()
            trait_value_min = trait_value.min()

            if trait_value_min == trait_value_max:
                return

            # Directly obtain descending index
            trait_value_sort_index = np.argsort(trait_value).astype(int)
            trait_value_sort_index = trait_value_sort_index[::-1]

            # Calculate the number of cells with>0
            _gt0_cell_size = (trait_value > 0).sum()

            _seed_cell_size = self._get_seed_cell_size_(_gt0_cell_size)

            seed_cell_count[i] = _seed_cell_size
            seed_cell_threshold[i] = trait_value[trait_value_sort_index[_seed_cell_size]]

            # Set seed cell index and weight
            _seed_cell_index = trait_value_sort_index[:_seed_cell_size]
            seed_cell_index[_seed_cell_index, i] = 1
            seed_cell_weight[_seed_cell_index, i] = self._get_seed_cell_weight_(
                seed_cell_index=_seed_cell_index, value=trait_value
            )

            # Enrichment interval index
            _enrichment_start = _seed_cell_size
            _enrichment_end = min(2 * _seed_cell_size, self.cell_size - 1)

            if _gt0_cell_size == _seed_cell_size:
                _enrichment_start = max(_seed_cell_size - self._enrichment_seed_cell_min_count_, 0)
                _enrichment_end = _seed_cell_size

            _seed_cell_en_index = trait_value_sort_index[_enrichment_start:_enrichment_end]
            seed_cell_weight_en[_seed_cell_en_index, i] = self._get_seed_cell_weight_(
                seed_cell_index=_seed_cell_index if len(_seed_cell_en_index) == len(_seed_cell_index) else _seed_cell_en_index,
                value=trait_value,
                seed_cell_index_enrichment=_seed_cell_en_index
            )

            if not self.is_simple and self.is_ablation:
                seed_cell_value = np.zeros(n_cells)
                seed_cell_value[_seed_cell_index] = 1
                seed_cell_matrix[:, i] = seed_cell_value / (1 if seed_cell_value.sum() == 0 else seed_cell_value.sum())

                seed_cell_en_value = np.zeros(n_cells)
                seed_cell_en_value[_seed_cell_en_index] = 1
                seed_cell_matrix_en[:, i] = seed_cell_en_value / (1 if seed_cell_en_value.sum() == 0 else seed_cell_en_value.sum())

        # Parallel processing of all traits and real-time display of progress
        Parallel(n_jobs=self.n_jobs, backend='threading')(
            delayed(_process_single_trait)(i) for i in tqdm(self.trait_range, desc="Obtain progress of seed cells with weights")
        )

        return seed_cell_count, seed_cell_threshold, seed_cell_matrix, seed_cell_weight, seed_cell_index, seed_cell_matrix_en, seed_cell_weight_en

    @staticmethod
    def scale_norm(score: matrix_data, is_verbose: bool = False) -> matrix_data:
        return trs_scale_norm(score, axis=0, is_verbose=is_verbose)

    def _simple_error_(self) -> None:

        if self.is_simple and "is_simple" in self.trs_adata.uns.keys() and self.trs_adata.uns["is_simple"]:
            ul.log(__name__).error("The parameter `is_simple` is True, so running this method is not supported.")
            raise RuntimeError("The parameter `is_simple` is True, so running this method is not supported.")

    def run_benchmark(self) -> None:
        """
        Perform random walk of random seeds on all traits.
        """
        self._simple_error_()

        ul.log(__name__).info(f"Calculate {len(self.trait_list)} traits/diseases for process `run_benchmark` (Count: {self.benchmark_count}). (Randomly perturb seed cells. ===> `benchmark`)")

        total_steps = len(self.trait_list) * self.benchmark_count
        with tqdm(total=total_steps) as pbar:
            for i in self.trait_range:

                random_seed_cell_matrix = np.zeros((self.cell_size, self.benchmark_count))

                # Obtain all cell score values in a trait
                trait_adata: AnnData = self.init_status[:, i]
                trait_value: collection = to_dense(trait_adata.X, is_array=True).flatten()

                # Obtain the maximum initial score
                trait_value_max = np.max(trait_value)
                trait_value_min = np.min(trait_value)

                for j in range(self.benchmark_count):
                    # Set random seed information
                    random_seed_cell = np.zeros(self.cell_size)
                    random_seed_index = np.random.choice(np.arange(0, self.cell_size), size=self.seed_cell_count[i], replace=False)

                    if trait_value_min != trait_value_max:
                        # seed cell weight
                        random_seed_cell[random_seed_index] = 1 / self.cell_size
                        random_seed_cell_matrix[:, j] = random_seed_cell

                    pbar.update(1)

                # Random walk
                cell_value_matrix = self._random_walk_core_(random_seed_cell_matrix)
                # Remove the influence of background
                self.random_seed_cell[:, i] = cell_value_matrix.mean(axis=1)

        cell_value = self.scale_norm(self.random_seed_cell)
        self.trs_adata.layers["benchmark"] = to_sparse(cell_value)
        self.is_benchmark = True

    @staticmethod
    def _get_label_description_(label: str) -> Tuple[str, str]:
        if label == "run_core" or label == "run_en":
            return "Calculate random walk with weighted seed cells.", "trs"
        elif label == "run_ablation_ncsw" or label == "run_en_ablation_ncsw":
            return "Removed cell weights in random walk and cluster type weights in initial scores.", "trs_ncsw"
        elif label == "run_ablation_nsw" or label == "run_en_ablation_nsw":
            return "Removed cell weights from random walk.", "trs_nsw"
        elif label == "run_ablation_ncw" or label == "run_en_ablation_ncw":
            return "Removed cell cluster type weights in initial scores.", "trs_ncw"
        elif label == "run_ablation_m_knn" or label == "run_en_ablation_m_knn":
            return "Using the M-KNN method during the execution of weighted random walks.", "trs_m_knn"
        elif label == "run_knock (positive)":
            return "Run knockout or knockdown by random walk with weight. (positive)", "knock_effect_positive"
        elif label == "run_knock (negative)":
            return "Run knockout or knockdown by random walk with weight. (negative)", "knock_effect_negative"
        elif label == "run_knock_control (control & positive)":
            return "Run knockout or knockdown by random walk with weight. (control & positive)", "knock_effect_positive_control"
        elif label == "run_knock_control (control & negative)":
            return "Run knockout or knockdown by random walk with weight. (control & negative)", "knock_effect_negative_control"
        else:
            raise ValueError(f"{label} is not a valid information.")

    def _run_(self, seed_cell_data: matrix_data, label: str, weight: matrix_data = None) -> matrix_data:
        """
        Calculate random walk
        :param seed_cell_data: Seed cell data
        :return: Return values without `scale` normalization
        """

        if weight is None:
            weight = self.weight

        _log_info_, _layer_label_ = self._get_label_description_(label)
        ul.log(__name__).info(f"Calculate {len(self.trait_list)} traits/diseases for process `{label}`. ({_log_info_} ===> `{_layer_label_}`)")

        score = self._random_walk_core_(seed_cell_data, weight=weight)

        ul.log(__name__).info("Normalize the results")
        cell_value = self.scale_norm(score)

        if _layer_label_ == "trs":
            self.trs_adata.X = to_sparse(cell_value)
        else:
            self.trs_adata.layers[_layer_label_] = to_sparse(cell_value)

        return score

    def run_core(self) -> None:
        """
        Calculate weighted random walk
        """
        if not self.is_simple:
            self.trs_adata.layers["seed_cell_weight"] = to_sparse(self.seed_cell_weight)

        self.trs_adata.layers["seed_cell_index"] = to_sparse(self.seed_cell_index)

        self.trs_adata.var["seed_cell_count"] = self.seed_cell_count
        self.trs_adata.var["seed_cell_threshold"] = self.seed_cell_threshold
        self.trs_source = self._run_(self.seed_cell_weight, "run_core")

        self.trs_adata.layers["trs_source"] = to_sparse(self.trs_source)
        self.is_run_core = True

    def run_ablation_m_knn(self) -> None:
        """
        Using M-KNN fully connected cellular network
        """
        self._simple_error_()
        self.trs_m_knn_source = self._run_(self.seed_cell_weight, "run_ablation_m_knn", self.weight_m_knn)
        self.is_run_ablation_m_knn = True

    def run_ablation_ncw(self) -> None:
        """
        Removed cell cluster type weights in initial scores
        """
        self._simple_error_()
        self.trs_adata.layers["seed_cell_weight_ncw"] = self.seed_cell_weight_ncw

        if "seed_cell_count_nw" not in self.trs_adata.var.columns:
            self.trs_adata.var["seed_cell_count_nw"] = self.seed_cell_count_nw

        if "seed_cell_threshold_nw" not in self.trs_adata.var.columns:
            self.trs_adata.var["seed_cell_threshold_nw"] = self.seed_cell_threshold_nw

        self.trs_ncw_source = self._run_(self.seed_cell_weight_ncw, "run_ablation_ncw")
        self.is_run_ablation_ncw = True

    def run_ablation_nsw(self) -> None:
        """
        Removed cell weights from random walk
        """
        self._simple_error_()
        self.trs_adata.layers["seed_cell_weight_nsw"] = self.seed_cell_weight_nsw
        self.trs_nsw_source = self._run_(self.seed_cell_weight_nsw, "run_ablation_nsw")
        self.is_run_ablation_nsw = True

    def run_ablation_ncsw(self) -> None:
        """
        Removed cell weights in random walk and cluster type weights in initial scores
        """
        self._simple_error_()
        self.trs_adata.layers["seed_cell_weight_ncsw"] = self.seed_cell_weight_ncsw

        if "seed_cell_count_nw" not in self.trs_adata.var.columns:
            self.trs_adata.var["seed_cell_count_nw"] = self.seed_cell_count_nw

        if "seed_cell_threshold_nw" not in self.trs_adata.var.columns:
            self.trs_adata.var["seed_cell_threshold_nw"] = self.seed_cell_threshold_nw

        self.trs_ncsw_source = self._run_(self.seed_cell_weight_ncsw, "run_ablation_ncsw")
        self.is_run_ablation_ncsw = True

    def _run_enrichment_(self, seed_cell_en_weight: matrix_data, label: str) -> None:
        """
        Enrichment analysis of traits/cells
        :param seed_cell_en_weight: Seed cell data
        """

        _layer_label_: str = "tre"

        source_value: matrix_data = self.trs_source

        _, _trs_layer_label_ = self._get_label_description_(label)

        if label == "run_en":
            if not self.is_run_core:
                ul.log(__name__).warning("Need to run the `run_core` method first in order to run this method. Start run...")
                self.run_core()

        elif label == "run_en_ablation_m_knn":
            if not self.is_run_ablation_m_knn:
                ul.log(__name__).warning("Need to run the `run_ablation_m_knn` method first in order to run this method. Start run...")
                self.run_ablation_m_knn()

            _layer_label_ = "tre_m_knn"
            source_value = self.trs_m_knn_source

        elif label == "run_en_ablation_ncw":
            if not self.is_run_ablation_ncw:
                ul.log(__name__).warning("Need to run the `run_ablation_ncw` method first in order to run this method. Start run...")
                self.run_ablation_ncw()

            _layer_label_ = "tre_ncw"
            source_value = self.trs_ncw_source

        elif label == "run_en_ablation_nsw":
            if not self.is_run_ablation_nsw:
                ul.log(__name__).warning("Need to run the `run_ablation_nsw` method first in order to run this method. Start run...")
                self.run_ablation_nsw()

            _layer_label_ = "tre_nsw"
            source_value = self.trs_nsw_source

        elif label == "run_en_ablation_ncsw":
            if not self.is_run_ablation_ncsw:
                ul.log(__name__).warning("Need to run the `run_ablation_ncsw` method first in order to run this method. Start run...")
                self.run_ablation_ncsw()

            _layer_label_ = "tre_ncsw"
            source_value = self.trs_ncsw_source

        else:
            raise ValueError(f"{label} error. `run_en`, `run_en_ablation_m_knn`, `run_en_ablation_ncw`, `run_en_ablation_nsw` or `run_en_ablation_ncsw`")

        cell_anno: DataFrame = self.cell_anno.copy()
        trs_score = to_dense(self.trs_adata.X if label == "run_en" else self.trs_adata.layers[_trs_layer_label_], is_array=True)

        # Initialize enriched container
        trait_cell_enrichment = np.zeros(self.trs_adata.shape).astype(int)
        trait_cell_credible = np.zeros(self.trs_adata.shape).astype(np.float32)

        ul.log(__name__).info(f"Calculate {len(self.trait_list)} traits/diseases for process `{label}`. (Enrichment-random walk)")
        # Random walk
        cell_value_data = self._random_walk_(
            seed_cell_en_weight,
            weight=self.weight_m_knn if label == "run_en_ablation_m_knn" else self.weight,
            gamma=self.enrichment_gamma
        )

        ul.log(__name__).info(f"Calculate {len(self.trait_list)} traits/diseases for process `{label}`. (Enrichment-score)")

        # Process each trait in parallel
        def _process_trait(i):
            # Random walk
            cell_value = cell_value_data[:, i]

            # separate
            cell_value_credible = mean_symmetric_scale(np.array(source_value[:, i]).flatten() - np.array(cell_value).flatten(), is_verbose=False)

            # This step is only executed if it contains cell clustering type weights
            if label == "run_en" or label == "run_en_ablation_nsw" or label == "run_en_ablation_m_knn":
                _enrichment_index_ = trs_score[:, i] > self.enrichment_threshold

                if np.any(_enrichment_index_):
                    # Ratio of cell clustering types enriched by threshold
                    _, _clustering_map_ = self._get_seed_cell_clustering_weight_(_enrichment_index_)
                    _clustering_weight_ = cell_anno["clusters"].map(_clustering_map_)
                    _clustering_weight_ = mean_symmetric_scale(_clustering_weight_, is_verbose=False)
                    _clustering_weight_mean_ = _clustering_weight_.mean()
                    # Correction score
                    cell_value_credible += (_clustering_weight_ - _clustering_weight_mean_)

            trait_cell_enrichment[:, i][cell_value_credible > self.credible_threshold] = 1
            trait_cell_credible[:, i] = cell_value_credible

        # Process each trait in parallel, backend='threading' can effectively prevent the read-only parameter issue caused by copying in loky multi-process mode
        Parallel(n_jobs=self.n_jobs, backend='threading')(delayed(_process_trait)(i) for i in tqdm(self.trait_range))

        self.trs_adata.layers[_layer_label_] = to_sparse(trait_cell_enrichment)

        if not self.is_simple:
            self.trs_adata.layers[f"credible_{_layer_label_}"] = to_sparse(trait_cell_credible)

    def run_enrichment(self) -> None:
        """
        Enrichment analysis
        """
        self._run_enrichment_(self.seed_cell_weight_en, "run_en")
        self.is_run_enrichment = True

    def run_en_ablation_m_knn(self) -> None:
        """
        Using M-KNN fully connected cellular network (Enrichment analysis)
        """
        self._simple_error_()
        self._run_enrichment_(self.seed_cell_weight_en, "run_en_ablation_m_knn")
        self.is_run_en_ablation_m_knn = True

    def run_en_ablation_ncw(self) -> None:
        """
        Removed cell cluster type weights in initial scores
        """
        self._simple_error_()
        self._run_enrichment_(self.seed_cell_weight_en_ncw, "run_en_ablation_ncw")
        self.is_run_en_ablation_ncw = True

    def run_en_ablation_nsw(self) -> None:
        """
        Removed cell weights from random walk
        """
        self._simple_error_()
        self._run_enrichment_(self.seed_cell_weight_en_nsw, "run_en_ablation_nsw")
        self.is_run_en_ablation_nsw = True

    def run_en_ablation_ncsw(self) -> None:
        """
        Removed cell weights in random walk and cluster type weights in initial scores
        """
        self._simple_error_()
        self._run_enrichment_(self.seed_cell_weight_en_ncsw, "run_en_ablation_ncsw")
        self.is_run_en_ablation_ncsw = True

    def run_knock(self, trs: AnnData, knock_trait: str, is_control: bool = False) -> None:

        if trs.shape[0] != self.cell_size:
            ul.log(__name__).error(f"The number of cells ({trs.shape[0]}) in the input `trs` is inconsistent with the number of cells ({self.cell_size}) in the knockdown after knockout")
            raise ValueError(f"The number of cells ({trs.shape[0]}) in the input `trs` is inconsistent with the number of cells ({self.cell_size}) in the knockdown after knockout.")

        if "trs_source" not in trs.layers:
            ul.log(__name__).error("`trs_source` is not in `trs.layers`, please execute function `ml.core` first to obtain the result as input for the `trs` parameter.")
            raise ValueError("`trs_source` is not in `trs.layers`, please execute function `ml.core` first to obtain the result as input for the `trs` parameter.")

        if "seed_cell_index" not in trs.layers:
            ul.log(__name__).error("`seed_cell_index` is not in `trs.layers`, please execute function `ml.core` first to obtain the result as input for the `trs` parameter.")
            raise ValueError("`seed_cell_index` is not in `trs.layers`, please execute function `ml.core` first to obtain the result as input for the `trs` parameter.")

        if knock_trait not in trs.var["id"]:
            ul.log(__name__).error(f"`{knock_trait}` trait or disease does not exist.")
            raise ValueError(f"`{knock_trait}` trait or disease does not exist.")

        knock_info_content = "run_knock_control" if is_control else "run_knock"

        ul.log(__name__).info(f"Calculate {len(self.trait_list)} for seed cells information. ({knock_info_content})")
        init_trait_source_value: matrix_data = to_dense(trs[:, knock_trait].layers["init_trs"])
        init_trait_value: matrix_data = to_dense(self.trs_adata.layers["init_trs"])

        init_trait_positive_effect = init_trait_source_value - init_trait_value
        init_trait_positive_effect[init_trait_positive_effect < 0] = 0
        init_trait_negative_effect = init_trait_value - init_trait_source_value
        init_trait_negative_effect[init_trait_negative_effect < 0] = 0

        self.trs_adata.layers["init_trait_positive"] = to_sparse(init_trait_positive_effect)
        self.trs_adata.layers["init_trait_negative"] = to_sparse(init_trait_negative_effect)

        init_trait_positive_adata = check_adata_get(self.trs_adata, "init_trait_positive")
        (_, positive_seed_cell_threshold, _, positive_seed_cell_weight, _, _, _,) = self._get_seed_cell_(init_data=init_trait_positive_adata, info="knock (positive)")

        init_trait_negative_adata = check_adata_get(self.trs_adata, "init_trait_negative")
        (_, negative_seed_cell_threshold, _, negative_seed_cell_weight, _, _, _,) = self._get_seed_cell_(init_data=init_trait_negative_adata, info="knock (negative)")

        self.trs_adata.var["positive_seed_cell_threshold"] = positive_seed_cell_threshold
        self.trs_adata.var["negative_seed_cell_threshold"] = negative_seed_cell_threshold

        if is_control:
            ul.log(__name__).info("Perturb the initialization TRS.")
            for i in tqdm(self.trait_range):
                positive_seed_cell_weight[:, i] = perturb_data(positive_seed_cell_weight[:, i], 1.0)
                negative_seed_cell_weight[:, i] = perturb_data(negative_seed_cell_weight[:, i], 1.0)

        # Obtain the result after random walk
        _positive_label_: str = "control & positive" if is_control else "positive"
        self.trs_source_positive = self._run_(positive_seed_cell_weight, f"{knock_info_content} ({_positive_label_})")
        _negative_label_: str = "control & negative" if is_control else "negative"
        self.trs_source_negative = self._run_(negative_seed_cell_weight, f"{knock_info_content} ({_negative_label_})")

        self.trs_adata.layers["knock_effect_positive_control" if is_control else "knock_effect_positive"] = to_sparse(self.trs_source_positive)
        self.trs_adata.layers["knock_effect_negative_control" if is_control else "knock_effect_negative"] = to_sparse(self.trs_source_negative)

        ul.log(__name__).info("Obtain the effect size of knocking out or knocking down ==> .layers[\"{}\"]".format("knock_effect_control" if is_control else "knock_effect"))
        knock_effect_value = self.trs_source_positive - self.trs_source_negative
        self.trs_adata.layers["knock_effect_source_control" if is_control else "knock_effect_source"] = to_sparse(knock_effect_value)
        self.trs_adata.layers["knock_effect_control" if is_control else "knock_effect"] = to_sparse(mean_symmetric_scale(knock_effect_value, axis=0, is_verbose=False))



# -*- coding: UTF-8 -*-

import os
from typing import Tuple, Union, Any

from pandas import DataFrame
import seaborn as sns

from .. import util as ul
from ..util import path, plot_end, plot_start

__name__: str = "plot_box"


def box_base(
    df: DataFrame,
    value: str = "value",
    x_name: str = None,
    y_name: str = "value",
    clusters: str = "clusters",
    palette: Union[Tuple, list] = None,
    width: float = 2,
    height: float = 2,
    bottom: float = 0.3,
    line_width: float = 0.3,
    marker_size: float = 0.2,
    rotation: float = 65,
    orient: str = None,
    title: str = None,
    whis: float = 1.5,
    show_fliers: bool = True,
    is_sort: bool = True,
    order_names: list = None,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:

    # judge
    df_columns = list(df.columns)

    if value not in df_columns:
        ul.log(__name__).error(f"The `value` ({value}) parameter must be in the `df` parameter data column name ({df_columns})")
        raise ValueError(f"The `value` ({value}) parameter must be in the `df` parameter data column name ({df_columns})")

    fig, ax = plot_start(width, height, bottom, output, show)

    group_columns = [clusters]

    new_df: DataFrame = df.groupby(group_columns, as_index=False)[value].median()

    if "color" in df_columns:
        new_df_color: DataFrame = df.groupby(group_columns, as_index=False)["color"].first()
        new_df = new_df.merge(new_df_color, how="left", on=clusters)

    colors: list = []

    # sort
    if is_sort:
        new_df.sort_values([value], ascending=False, inplace=True)
        y_names: Union[list, None] = list(new_df[clusters])

        if "color" in df_columns:
            colors = list(new_df["color"])

    else:
        new_df.index = new_df[clusters]

        if order_names is not None:
            y_names: list = order_names

            if "color" in df_columns:

                for i in order_names:

                    for j, c in zip(new_df[clusters], new_df["color"]):

                        if i == j:
                            colors.append(c)
                            break

        else:
            y_names = new_df[clusters]

            if "color" in df_columns:
                colors = list(new_df["color"])

    # scatter
    sns.boxplot(
        data=df,
        x=clusters,
        y=value,
        order=y_names,
        showfliers=show_fliers,
        fliersize=marker_size,
        orient=orient,
        whis=whis,
        ax=ax,
        flierprops={'marker': 'o', 'markersize': marker_size},
        boxprops={'linestyle': '-', 'linewidth': line_width},
        whiskerprops={'linestyle': '-', 'linewidth': line_width},
        medianprops={'linestyle': '-', 'linewidth': line_width},
        palette=palette if palette is not None else (colors if "color" in df_columns else None),
        **kwargs
    )

    lines = ax.lines

    for line in lines:
        line.set_linewidth(line_width)

    # set coordinate
    ax.set_xticks(range(len(y_names)))
    ax.set_xticklabels(labels=y_names, rotation=rotation)
    ax.spines['top'].set_linewidth(line_width)
    ax.spines['bottom'].set_linewidth(line_width)
    ax.spines['left'].set_linewidth(line_width)
    ax.spines['right'].set_linewidth(line_width)

    ax.yaxis.grid(True, linestyle='-', linewidth=line_width)

    plot_end(fig, title, x_name, y_name, output, show, close)


def box_trait(
    trait_df: DataFrame,
    trait_name: str = "All",
    trait_column_name: str = "id",
    value: str = "value",
    clusters: str = "clusters",
    x_name: str = None,
    y_name: str = "value",
    palette: Tuple = None,
    orient: str = None,
    width: float = 2,
    height: float = 2,
    line_width: float = 0.1,
    marker_size: float = 0.5,
    bottom: float = 0.3,
    rotation: float = 65,
    whis: float = 1.5,
    show_fliers: bool = True,
    is_sort: bool = True,
    order_names: list = None,
    title: str = None,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:

    data: DataFrame = trait_df.copy()

    def trait_plot(trait_: str, atac_cell_df_: DataFrame) -> None:
        """
        show plot
        :param trait_: trait name
        :param atac_cell_df_:
        :return: None
        """
        ul.log(__name__).info("Plotting box {}".format(trait_))
        # get gene score
        trait_score = atac_cell_df_[atac_cell_df_[trait_column_name] == trait_]
        # Sort gene scores from small to large
        box_base(
            df=trait_score,
            value=value,
            x_name=x_name,
            y_name=y_name,
            width=width,
            palette=palette,
            height=height,
            bottom=bottom,
            rotation=rotation,
            is_sort=is_sort,
            whis=whis,
            order_names=order_names,
            line_width=line_width,
            show_fliers=show_fliers,
            marker_size=marker_size,
            orient=orient,
            clusters=clusters,
            title=f"{title} {trait_}" if title is not None else title,
            output=os.path.join(output, f"cell_{trait_}_score_box.pdf") if output is not None else None,
            show=show,
            close=close,
            **kwargs
        )

    # noinspection DuplicatedCode
    trait_list = list(set(data[trait_column_name]))
    # judge trait
    if trait_name != "All" and trait_name not in trait_list:
        ul.log(__name__).error(f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}.")
        raise ValueError(f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}.")

    # plot
    if trait_name == "All":
        for trait in trait_list:
            trait_plot(trait, trait_df)
    else:
        trait_plot(trait_name, trait_df)



# -*- coding: UTF-8 -*-

import os
import shutil
import warnings
from typing import Optional, Union, List, Literal

import pandas as pd
from pandas import DataFrame
from anndata import AnnData
import numpy as np

from ._scanpy_ import filter_data

from .. import util as ul
from ..file import read_sc_atac, read_h5ad, save_h5ad
from ..util import path, collection, set_inf_value, chrtype, add_cluster_info, generate_str

__name__: str = "preprocessing_snapatac2"


def get_feature_count(raw_count: int, need_features: Optional[Union[int | float]]) -> int:
    return int(raw_count * need_features) if need_features <= 1 else int(need_features)


def _process_sc_atac_(
    fragment_file: path | List[path],
    genome_anno,
    h5ad_file: Optional[path | List[path]] = None,
    min_num_fragments: int = 200,
    sorted_by_barcode: bool = False,
    bin_size: int = 500,
    min_tsse: float = 5.0,
    counting_strategy: Literal['fragment', 'insertion', 'paired-insertion'] = 'paired-insertion',
    is_filter_doublets: bool = True,
    need_features: Optional[Union[int | float]] = None
):
    ul.log(__name__).info(f"Read {fragment_file}")

    import snapatac2 as snap

    data = snap.pp.import_data(
        fragment_file,
        chrom_sizes=genome_anno,
        file=h5ad_file,
        min_num_fragments=min_num_fragments,
        sorted_by_barcode=sorted_by_barcode
    )

    if isinstance(fragment_file, path):
        ul.log(__name__).info(f"Read finish, shape: {data.shape}")
    else:
        ul.log(__name__).info(f"Read finish")

    # the standard procedures to add tile matrices, select features, and identify doublets
    snap.metrics.tsse(data, genome_anno)

    if isinstance(fragment_file, path):
        ul.log(__name__).info(f"Shape: {data.shape}")

    ul.log(__name__).info(f"Filter cells through TSSE ({min_tsse})")
    snap.pp.filter_cells(data, min_tsse=min_tsse)

    if isinstance(fragment_file, path):
        ul.log(__name__).info(f"Shape: {data.shape}")

    ul.log(__name__).info(f"Add tile matrix.")
    snap.pp.add_tile_matrix(data, bin_size=bin_size, counting_strategy=counting_strategy)

    if isinstance(fragment_file, path):
        ul.log(__name__).info(f"Shape: {data.shape}")

    # The situation where features have been filtered
    features: Optional[str] = None

    if need_features is not None:
        feature_count: int = get_feature_count(
            data.shape[1] if isinstance(fragment_file, path) else data[0].shape[1], need_features
        )
        ul.log(__name__).info("Select {} features".format(feature_count))
        snap.pp.select_features(data, n_features=feature_count)

        if isinstance(fragment_file, path):
            ul.log(__name__).info(f"Shape: {data.shape}")

        features = "selected"

    if is_filter_doublets:
        ul.log(__name__).info("Identify doublets.")
        snap.pp.scrublet(data, features=features)
        ul.log(__name__).info("Filter doublets.")
        snap.pp.filter_doublets(data)

        if isinstance(fragment_file, path):
            ul.log(__name__).info(f"Shape: {data.shape}")

    _params_: dict = {
        "fragment_file": fragment_file,
        "h5ad_file": h5ad_file if h5ad_file is not None else "",
        "min_num_fragments": min_num_fragments,
        "sorted_by_barcode": sorted_by_barcode,
        "bin_size": bin_size,
        "min_tsse": min_tsse,
        "features": features if features is not None else "",
        "need_features": need_features if need_features is not None else "",
        "is_filter_doublets": is_filter_doublets
    }

    if not isinstance(data, list):
        data.uns["params"] = _params_

    return data, features


def get_sc_atac(
    fragment_file: path,
    genome_anno,
    h5ad_file: Optional[path] = None,
    min_num_fragments: int = 200,
    sorted_by_barcode: bool = False,
    bin_size: int = 500,
    min_tsse: float = 5.0,
    counting_strategy: Literal['fragment', 'insertion', 'paired-insertion'] = 'paired-insertion',
    need_features: Optional[Union[int | float]] = None,
    is_filter_doublets: bool = True
):

    features: Optional[str] = None

    if fragment_file.endswith(".h5ad"):
        h5ad_file = fragment_file

    if h5ad_file is not None and os.path.exists(h5ad_file):
        ul.log(__name__).warning(
            "Suggest using fragments files to scATAC-seq data. (If the reading process is completed through snapATAC2, "
            "this message can be ignored.)"
        )
        adata = read_sc_atac(h5ad_file)
    else:

        _cache_file_path_ = os.path.join(ul.project_cache_path, generate_str())
        _is_cache_ = False

        if h5ad_file is None:
            _is_cache_ = True
            ul.file_method(__name__).makedirs(_cache_file_path_)
            h5ad_file = os.path.join(_cache_file_path_, generate_str() + ".h5ad")

        # import the fragment files and process them
        adata, features = _process_sc_atac_(
            fragment_file=fragment_file,
            genome_anno=genome_anno,
            h5ad_file=h5ad_file,
            min_num_fragments=min_num_fragments,
            sorted_by_barcode=sorted_by_barcode,
            bin_size=bin_size,
            min_tsse=min_tsse,
            counting_strategy=counting_strategy,
            is_filter_doublets=is_filter_doublets,
            need_features=need_features
        )
        adata.close()

        adata = read_sc_atac(h5ad_file)

        if _is_cache_ and os.path.exists(_cache_file_path_):
            shutil.rmtree(_cache_file_path_)

        del _cache_file_path_, _is_cache_

    selected_list = None

    if features is not None:
        selected_list = np.array(list(adata.var["selected"]))

    return adata, selected_list, h5ad_file


def merge_sc_atac(
    files: dict,
    genome_anno,
    merge_key: str = "merge_sc_atac",
    min_num_fragments: int = 200,
    sorted_by_barcode: bool = False,
    bin_size: int = 500,
    min_tsse: float = 5.0,
    counting_strategy: Literal['fragment', 'insertion', 'paired-insertion'] = 'paired-insertion',
    max_iter_harmony: int = 20,
    harmony_groupby: Optional[Union[str | list[str]]] = None,
    is_selected: bool = False,
    is_batch: bool = True,
    need_features: Optional[Union[int | float]] = None,
    output_path: Optional[path] = None
) -> AnnData:
    """
    Integrate multiple scATAC-seq data through snapATAC2 (https://kzhang.org/SnapATAC2/tutorials/integration.html)
        Note: Please do not move the generated files during this processing
    :param files: {file_key: file path of scATAC-seq data, ...} scATAC-seq data information that needs to be integrated
    :param genome_anno: Reference genome, commonly known as `snap.genome.hg38` and `snap.genome.hg19`
    :param merge_key: Finally form the file name of H5AD
    :param min_num_fragments: Number of unique fragments threshold used to filter cells
    :param sorted_by_barcode: Is the input fragments file sorted
    :param bin_size: The size of consecutive genomic regions used to record the counts.
    :param min_tsse: Minimum TSS enrichment score required for a cell to pass filtering.
    :param max_iter_harmony: The maximum number of iterations in the `harmony` algorithm.
    :param harmony_groupby: If specified, split the data into groups and perform batch correction on each group separately.
    :param is_selected: If True, based on the feature selection in the `snap.pp.select_features` method, further
        filtering is performed according to the features of each sample.
    :param is_batch: If True, batch correction by sample.
    :param need_features: If `need_features` <=1, it represents the retention of `need_features`% of the overall
        features. Otherwise, it is considered an integer and `need_features` features are filtered.
    :param output_path: Path to generate file
    :return: Integrated scATAC-seq data
    """
    ul.log(__name__).info("Start integrating scATAC-seq data.")

    import snapatac2 as snap

    # Obtain sample information
    filenames: list = list(files.keys())

    if len(filenames) <= 1:
        ul.log(__name__).error("At least two samples are required.")
        raise ValueError("At least two samples are required.")

    # file path
    file_list: list = list(files.values())

    # The situation where features have been filtered
    features: Optional[str] = None

    if str(file_list[0]).endswith(".h5ad"):
        ul.log(__name__).warning(
            "Suggest using fragments files to integrate scATAC-seq data. (If the process of adding tile matrices, "
            "selecting features, and identifying doubles has been completed through snapATAC2, this message can be "
            "ignored.)"
        )
        adata_list: list = [(name, adata) for name, adata in zip(filenames, file_list)]
    else:
        output_filenames: list = [os.path.join(output_path, key + ".h5ad") for key in list(files.keys())]

        # import the fragment files and process them
        adatas, features = _process_sc_atac_(
            fragment_file=file_list,
            genome_anno=genome_anno,
            h5ad_file=output_filenames if output_path is not None else None,
            min_num_fragments=min_num_fragments,
            sorted_by_barcode=sorted_by_barcode,
            bin_size=bin_size,
            min_tsse=min_tsse,
            counting_strategy=counting_strategy,
            need_features=need_features
        )

        adata_list: list = [(name, adata) for name, adata in zip(filenames, adatas)]

        del adatas

    # AnnDataSet
    merge_filename: str = os.path.join(output_path, merge_key + ".h5ad") if output_path is not None else None
    data = snap.AnnDataSet(adatas=adata_list, filename=merge_filename)

    # id unique
    data.obs['barcodes'] = data.obs_names
    unique_cell_ids = [sa + '_' + bc for sa, bc in zip(data.obs['sample'], data.obs_names)]
    data.obs_names = unique_cell_ids

    selected_list: collection = np.array([])

    if need_features is not None:
        snap.pp.select_features(data, n_features=get_feature_count(data.shape[1], need_features))
        selected_list = np.array(list(data.var["selected"]))

    # spectral
    ul.log(__name__).info("Spectral dimensionality reduction.")
    snap.tl.spectral(data, features=features)

    # Batch correction
    if is_batch:
        ul.log(__name__).info("Batch correction.")
        snap.pp.mnc_correct(data, batch="sample")
        snap.pp.harmony(data, batch="sample", groupby=harmony_groupby, max_iter_harmony=max_iter_harmony)

    # close
    data.close()

    # read file
    sc_atac = read_sc_atac(merge_filename)

    # form count matrix
    data_matrix = []
    adata_path_list = list(sc_atac.uns["AnnDataSet"]["file_path"])

    for _adata_path_ in adata_path_list:
        _adata_ = read_h5ad(_adata_path_)
        data_matrix.append(_adata_.X)

        if is_selected and need_features is not None:
            selected_list = np.logical_and(selected_list, np.array(list(_adata_.var["selected"])))

        del _adata_

    from scipy.sparse import vstack

    sc_atac.X = vstack(data_matrix)

    if need_features is not None:
        sc_atac = sc_atac[:, selected_list]

    sc_atac.obs.index = sc_atac.obs.index.astype(str)
    sc_atac.var.index = sc_atac.var.index.astype(str)

    sc_atac.uns["params"] = {
        "files": files,
        "merge_key": merge_key,
        "min_num_fragments": min_num_fragments,
        "sorted_by_barcode": sorted_by_barcode,
        "bin_size": bin_size,
        "min_tsse": min_tsse,
        "max_iter_harmony": max_iter_harmony,
        "harmony_groupby": harmony_groupby,
        "is_selected": is_selected,
        "need_features": need_features,
        "output_path": output_path
    }

    if output_path is not None:
        save_h5ad(sc_atac, os.path.join(output_path, f"{merge_key}_snapATAC2.h5ad"))

    ul.log(__name__).info("End integrating scATAC-seq data.")
    return sc_atac


def get_gene_expression(
    fragment_file: path,
    genome_anno,
    h5ad_file: Optional[path] = None,
    min_num_fragments: int = 200,
    sorted_by_barcode: bool = False,
    bin_size: int = 500,
    min_tsse: float = 5.0,
    need_features: Optional[Union[int | float]] = None,
    min_cells: int = 5,
    is_filter_doublets: bool = True,
    gene_save_file: Optional[path] = None
) -> AnnData:

    import snapatac2 as snap

    # import the fragment files and process them
    adata, selected_list, h5ad_file = get_sc_atac(
        fragment_file=fragment_file,
        genome_anno=genome_anno,
        h5ad_file=h5ad_file,
        min_num_fragments=min_num_fragments,
        sorted_by_barcode=sorted_by_barcode,
        bin_size=bin_size,
        min_tsse=min_tsse,
        is_filter_doublets=is_filter_doublets,
        need_features=need_features
    )

    # Create the cell by gene activity matrix
    ul.log(__name__).info("Obtain gene expression matrix.")
    selected_adata = adata[:, selected_list] if selected_list is not None else adata
    filter_data(selected_adata)
    gene_matrix = snap.pp.make_gene_matrix(selected_adata, genome_anno)
    ul.log(__name__).info(f"Shape: {selected_adata.shape}")

    import scanpy as sc

    # normalize
    ul.log(__name__).info(f"Filter genes. shape: {gene_matrix.shape}")
    sc.pp.filter_genes(gene_matrix, min_cells=min_cells)
    ul.log(__name__).info(f"Gene matrix shape: {gene_matrix.shape}")

    ul.log(__name__).info(f"Normalize, log1p and magic.")
    sc.pp.normalize_total(gene_matrix)
    sc.pp.log1p(gene_matrix)
    sc.external.pp.magic(gene_matrix, solver="approximate")

    save_h5ad(adata, h5ad_file)

    gene_matrix.uns["params"] = {
        "fragment_file": fragment_file,
        "h5ad_file": h5ad_file,
        "min_num_fragments": min_num_fragments,
        "sorted_by_barcode": sorted_by_barcode,
        "bin_size": bin_size,
        "min_tsse": min_tsse,
        "need_features": need_features,
        "is_filter_doublets": is_filter_doublets,
        "gene_save_file": gene_save_file
    }

    if gene_save_file is not None:
        save_h5ad(gene_matrix, gene_save_file)

    return gene_matrix


def get_peak_matrix(
    fragment_file: path,
    genome_anno,
    cluster: str,
    cell_anno: Optional[DataFrame] = None,
    h5ad_file: Optional[path] = None,
    min_num_fragments: int = 200,
    sorted_by_barcode: bool = False,
    bin_size: int = 500,
    min_tsse: float = 5.0,
    need_features: Optional[Union[int | float]] = None,
    is_filter_doublets: bool = True,
    peak_matrix_save_file: Optional[path] = None
):

    import snapatac2 as snap

    # import the fragment files and process them
    adata, selected_list, h5ad_file = get_sc_atac(
        fragment_file=fragment_file,
        genome_anno=genome_anno,
        h5ad_file=h5ad_file,
        min_num_fragments=min_num_fragments,
        sorted_by_barcode=sorted_by_barcode,
        bin_size=bin_size,
        min_tsse=min_tsse,
        is_filter_doublets=is_filter_doublets,
        need_features=need_features
    )

    # add cell annotation information
    adata.obs = add_cluster_info(adata.obs, cell_anno, cluster)

    if cluster not in adata.obs.columns:
        ul.log(__name__).error(f"`{cluster}` is not in `adata.obs.columns`.")
        raise ValueError(f"`{cluster}` is not in `adata.obs.columns` ({adata.obs.columns}).")

    selected_adata = adata[:, selected_list] if selected_list is not None else adata
    filter_data(selected_adata)

    ul.log(__name__).info(f"Peak calling at the `{cluster}`-level.")
    try:
        snap.tl.macs3(selected_adata, groupby=cluster)
    except Exception as e:
        ul.log(__name__).error(f"The `cluster`({cluster}) is likely to have `NaN` values, please check.")
        raise RuntimeError(f"The `cluster`({cluster}) is likely to have `NaN` values, please check.\n {e}")

    ul.log(__name__).info("Obtain a unified, non-overlapping, and fixed-width peak list.")
    peaks = snap.tl.merge_peaks(selected_adata.uns['macs3'], genome_anno)

    ul.log(__name__).info("Create a cell by peak matrix.")
    peak_mat = snap.pp.make_peak_matrix(selected_adata, use_rep=peaks['Peaks'])

    peak_mat.uns["params"] = {
        "fragment_file": fragment_file,
        "cluster": cluster,
        "h5ad_file": h5ad_file,
        "min_num_fragments": min_num_fragments,
        "sorted_by_barcode": sorted_by_barcode,
        "bin_size": bin_size,
        "min_tsse": min_tsse,
        "need_features": need_features,
        "is_filter_doublets": is_filter_doublets,
        "peak_matrix_save_file": peak_matrix_save_file
    }

    if peak_matrix_save_file is not None:
        save_h5ad(peak_mat, peak_matrix_save_file)

    return adata, selected_adata, h5ad_file, peaks, peak_mat


def _process_info_to_adata_(
    adata: AnnData,
    selected_adata: AnnData,
    cluster: str,
    info_data: DataFrame,
    obs_info: DataFrame,
    h5ad_file: str
) -> AnnData:
    obs_unique = list(obs_info.index)
    obs_unique_dict: dict = dict(zip(obs_unique, range(len(obs_unique))))
    # var
    cluster_info: DataFrame = selected_adata.obs.groupby(cluster, as_index=False).size()
    cluster_info.index = cluster_info[cluster].astype(str)
    cluster_info.rename_axis("index", inplace=True)
    cluster_info.sort_values([cluster], inplace=True)
    cluster_list: list = list(cluster_info.index)

    shape = (len(obs_unique), len(cluster_list))
    ul.log(__name__).info(f"Create data, shape {shape}")
    log2_fold_change_matrix = np.zeros(shape)
    p_value_matrix = np.zeros(shape)
    adjusted_p_value_matrix = np.zeros(shape)

    # Add value
    for _id_, log2_fold_change, p_value, adjusted_p_value, _cluster_ in zip(
        info_data["id"],
        info_data["log2_fold_change"],
        info_data["p_value"],
        info_data["adjusted_p_value"],
        info_data[cluster]
    ):
        log2_fold_change_matrix[obs_unique_dict[_id_], cluster_list.index(_cluster_)] = log2_fold_change
        p_value_matrix[obs_unique_dict[_id_], cluster_list.index(_cluster_)] = p_value
        adjusted_p_value_matrix[obs_unique_dict[_id_], cluster_list.index(_cluster_)] = adjusted_p_value

    # solve -Inf/Inf value
    set_inf_value(log2_fold_change_matrix)
    set_inf_value(p_value_matrix)
    set_inf_value(adjusted_p_value_matrix)

    adjusted_p_value_matrix[adjusted_p_value_matrix == 0] = np.min(adjusted_p_value_matrix[adjusted_p_value_matrix != 0])

    # create
    info_adata: AnnData = AnnData(p_value_matrix, obs=obs_info, var=cluster_info)
    info_adata.layers["log2_fold_change"] = log2_fold_change_matrix
    info_adata.layers["adjusted_p_value"] = adjusted_p_value_matrix

    save_h5ad(adata, h5ad_file)

    return info_adata


def get_tf_data(
    fragment_file: path,
    genome_anno,
    cluster: str,
    cell_anno: Optional[DataFrame] = None,
    h5ad_file: Optional[path] = None,
    min_num_fragments: int = 200,
    sorted_by_barcode: bool = False,
    bin_size: int = 500,
    min_tsse: float = 5.0,
    need_features: Optional[Union[int | float]] = None,
    p_value: float = 0.01,
    is_filter_doublets: bool = True,
    peak_matrix_save_file: Optional[path] = None,
    tf_save_file: Optional[path] = None
) -> AnnData:

    import snapatac2 as snap

    # import the fragment files and process them
    adata, selected_adata, h5ad_file, peaks, peak_mat = get_peak_matrix(
        fragment_file=fragment_file,
        genome_anno=genome_anno,
        cluster=cluster,
        cell_anno=cell_anno,
        h5ad_file=h5ad_file,
        min_num_fragments=min_num_fragments,
        sorted_by_barcode=sorted_by_barcode,
        bin_size=bin_size,
        min_tsse=min_tsse,
        is_filter_doublets=is_filter_doublets,
        need_features=need_features,
        peak_matrix_save_file=peak_matrix_save_file
    )

    ul.log(__name__).info("Finding marker regions.")
    marker_peaks = snap.tl.marker_regions(peak_mat, groupby=cluster, pvalue=p_value)

    if len(marker_peaks.keys()) == 0 and p_value < 0.05:
        p_value = 0.05
        ul.log(__name__).warning(f"Due to the absence of marker regions, the `p_value` value is automatically changed to `0.05`.")
        marker_peaks = snap.tl.marker_regions(peak_mat, groupby=cluster, pvalue=p_value)

    if len(marker_peaks.keys()) == 0:
        ul.log(__name__).error(f"No marker regions for `p_value` = {p_value}, you can try to increase the `p_value`.")
        raise ValueError(f"No marker regions for `p_value` = {p_value}, you can try to increase the `p_value`.")

    ul.log(__name__).info("Motif enrichment.")

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        motifs = snap.tl.motif_enrichment(
            motifs=snap.datasets.cis_bp(unique=True),
            regions=marker_peaks,
            genome_fasta=genome_anno,
        )

    have_cluster_list: list = list(motifs.keys())
    ul.log(__name__).info(f"Merge motif result for `{have_cluster_list}`")

    motif_list: list = []
    columns: list = ['id', 'name', 'family', 'log2_fold_change', 'p_value', 'adjusted_p_value']

    # Add motif
    for cell_type in have_cluster_list:
        motif_data: DataFrame = pd.DataFrame(motifs[cell_type], columns=columns)
        motif_data[cluster] = cell_type
        motif_list.append(motif_data)

    # obtain all motif data
    tf_data: DataFrame = pd.concat(motif_list, axis=0)
    tf_data.drop(columns=["family"], inplace=True)

    # obs
    tf_info: DataFrame = tf_data[["id", "name"]].drop_duplicates()
    tf_info.index = tf_info["id"].astype(str)
    tf_info.rename_axis("index", inplace=True)
    tf_info.sort_values(["name", "id"], inplace=True)

    tf_adata = _process_info_to_adata_(
        adata=adata,
        selected_adata=selected_adata,
        cluster=cluster,
        info_data=tf_data,
        obs_info=tf_info,
        h5ad_file=h5ad_file
    )

    tf_adata.uns["params"] = {
        "fragment_file": fragment_file,
        "cluster": cluster,
        "h5ad_file": h5ad_file,
        "min_num_fragments": min_num_fragments,
        "sorted_by_barcode": sorted_by_barcode,
        "bin_size": bin_size,
        "min_tsse": min_tsse,
        "need_features": need_features,
        "p_value": p_value,
        "is_filter_doublets": is_filter_doublets,
        "peak_matrix_save_file": peak_matrix_save_file,
        "tf_save_file": tf_save_file
    }

    if tf_save_file is not None:
        save_h5ad(tf_adata, tf_save_file)

    return tf_adata


def get_difference_peaks(
    fragment_file: path,
    genome_anno,
    cluster: str,
    cell_anno: Optional[DataFrame] = None,
    h5ad_file: Optional[path] = None,
    min_num_fragments: int = 200,
    sorted_by_barcode: bool = False,
    bin_size: int = 500,
    min_tsse: float = 5.0,
    need_features: Optional[Union[int | float]] = None,
    is_filter_doublets: bool = True,
    min_log_fc: float = 0.25,
    min_pct: float = 0.05,
    peak_matrix_save_file: Optional[path] = None,
    diff_peaks_save_file: Optional[path] = None
) -> AnnData:

    import snapatac2 as snap

    # import the fragment files and process them
    adata, selected_adata, h5ad_file, peaks, peak_mat = get_peak_matrix(
        fragment_file=fragment_file,
        genome_anno=genome_anno,
        cluster=cluster,
        cell_anno=cell_anno,
        h5ad_file=h5ad_file,
        min_num_fragments=min_num_fragments,
        sorted_by_barcode=sorted_by_barcode,
        bin_size=bin_size,
        min_tsse=min_tsse,
        is_filter_doublets=is_filter_doublets,
        need_features=need_features,
        peak_matrix_save_file=peak_matrix_save_file
    )
    adata_cell_anno: DataFrame = selected_adata.obs.copy()
    cluster_list: list = list(set(adata_cell_anno[cluster]))
    cluster_list.sort()
    cluster_str: str = ", ".join(cluster_list)

    # Container for storing differential peaks
    diff_peaks_list: list = []

    # feature name, log2(fold_change), p-value, adjusted p-value
    columns: list = ["id", "log2_fold_change", "p_value", "adjusted_p_value"]

    for _cluster_ in cluster_list:
        ul.log(__name__).info(f"Processing {cluster}: {_cluster_}/({cluster_str}).")
        _cluster_list_ = cluster_list.copy()
        _cluster_list_.remove(_cluster_)

        # cell barcodes
        cell_group1 = adata_cell_anno[adata_cell_anno[cluster] == _cluster_].index
        cell_group2 = adata_cell_anno[adata_cell_anno[cluster].isin(_cluster_list_)].index

        ul.log(__name__).info(f"Processing {cluster}: {_cluster_} difference peaks.")

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            _diff_peaks_ = snap.tl.diff_test(
                peak_mat,
                cell_group1=cell_group1,
                cell_group2=cell_group2,
                min_log_fc=min_log_fc,
                min_pct=min_pct
            )

        # process result
        _diff_peaks_data_ = pd.DataFrame(_diff_peaks_, columns=columns)
        _diff_peaks_data_["log2_fold_change"] = _diff_peaks_data_["log2_fold_change"].astype(np.float64)
        _diff_peaks_data_["p_value"] = _diff_peaks_data_["p_value"].astype(np.float64)
        _diff_peaks_data_["adjusted_p_value"] = _diff_peaks_data_["adjusted_p_value"].astype(np.float64)
        _diff_peaks_data_.insert(0, cluster, _cluster_)
        _diff_peaks_data_.index = _diff_peaks_data_[cluster].astype(str) + "_" + _diff_peaks_data_["id"].astype(str)

        # Add result
        diff_peaks_list.append(_diff_peaks_data_)

        del _cluster_list_, _diff_peaks_, _diff_peaks_data_

    # get difference peaks data
    diff_peaks_data: DataFrame = pd.concat(diff_peaks_list, axis=0)

    # obs
    peaks_info: DataFrame = diff_peaks_data["id"].drop_duplicates().str.split(":|-", expand=True)
    peaks_info.columns = ["chr", "start", "end"]
    peaks_info = peaks_info[peaks_info["chr"].isin(list(chrtype.categories))]
    peaks_info["chr"] = peaks_info["chr"].astype(chrtype)
    peaks_info.index = (
        peaks_info["chr"].astype(str) + ":" + peaks_info["start"].astype(str) + "-" + peaks_info["end"].astype(str)
    )
    peaks_info.sort_values(["chr", "start", "end"], inplace=True)

    diff_peaks_adata = _process_info_to_adata_(
        adata=adata,
        selected_adata=selected_adata,
        cluster=cluster,
        info_data=diff_peaks_data,
        obs_info=peaks_info,
        h5ad_file=h5ad_file
    )

    diff_peaks_adata.uns["params"] = {
        "fragment_file": fragment_file,
        "cluster": cluster,
        "h5ad_file": h5ad_file,
        "min_num_fragments": min_num_fragments,
        "sorted_by_barcode": sorted_by_barcode,
        "bin_size": bin_size,
        "min_tsse": min_tsse,
        "need_features": need_features,
        "is_filter_doublets": is_filter_doublets,
        "min_log_fc": min_log_fc,
        "min_pct": min_pct,
        "peak_matrix_save_file": min_pct,
        "diff_peaks_save_file": diff_peaks_save_file
    }

    if diff_peaks_save_file is not None:
        save_h5ad(diff_peaks_adata, diff_peaks_save_file)

    return diff_peaks_adata



# -*- coding: UTF-8 -*-

import os.path
from typing import Union, Tuple, Optional, Any

import matplotlib
import numpy as np
import pandas as pd
from anndata import AnnData
from matplotlib import pyplot as plt
from matplotlib.colors import ListedColormap
from pandas import DataFrame
import seaborn as sns

from .. import util as ul
from ..util import path, collection, type_50_colors, type_20_colors, chrtype, type_set_colors, plot_end, plot_start

__name__: str = "plot_pie"

matplotlib.set_loglevel("error")


def scatter_base(
    df: DataFrame,
    x: str,
    y: str,
    hue: str = None,
    hue_order: list = None,
    x_name: str = None,
    y_name: str = None,
    title: str = None,
    bar_label: str = None,
    cmap: str = "Oranges",
    width: float = 2,
    height: float = 2,
    right: float = 0.9,
    bottom: float = 0,
    text_fontsize: float = 7,
    legend_fontsize: float = 7,
    start_color_index: int = 0,
    color_step_size: int = 0,
    type_colors: collection = None,
    edge_color: str = None,
    size: Union[float, collection] = 1.0,
    legend: dict = None,
    number: bool = False,
    is_text: bool = False,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:
    fig, ax = plot_start(width, height, bottom, output, show)

    # scatter
    if number:
        norm = plt.Normalize(df[hue].min(), df[hue].max())
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
        sm.set_array([])
        plt.colorbar(sm, label=bar_label)
        sns.scatterplot(
            data=df,
            x=x,
            y=y,
            palette=cmap,
            hue=hue,
            s=size,
            legend=False,
            edgecolor=edge_color,
            **kwargs
        )
    else:
        __hue_order__ = list(np.sort(list(set(df[hue]))))

        if type_colors is None:
            type_colors = type_20_colors if len(__hue_order__) <= 20 else type_50_colors

        colors = {}

        if legend is not None:
            df.loc[:, "__hue__"] = df[hue].copy()

        i = 0
        for elem in __hue_order__:
            if legend is not None:
                df.loc[df[df["__hue__"] == elem].index, "__hue__"] = legend[elem]
                colors.update(
                    {legend[elem]: type_colors[start_color_index + i * color_step_size + __hue_order__.index(elem)]}
                )
            else:
                colors.update(
                    {
                        elem: type_colors[start_color_index + i * color_step_size + __hue_order__.index(elem)]
                    }
                )
            i += 1

        if legend is not None:
            if hue_order is None:
                hue_order = list(np.sort(list(set(df["__hue__"]))))
        else:
            if hue_order is None:
                hue_order = __hue_order__

        sns.scatterplot(
            data=df,
            x=x,
            y=y,
            edgecolor=edge_color,
            palette=colors,
            hue="__hue__" if legend is not None else hue,
            hue_order=hue_order,
            s=size,
            **kwargs
        )

        if is_text:

            df_anno = df[[hue, x, y]].groupby(hue, as_index=False).mean()

            for txt, i, j in zip(df_anno[hue], df_anno[x], df_anno[y]):
                plt.annotate(
                    txt,
                    xy=(i, j),
                    xytext=(-10, 0),
                    textcoords="offset points",
                    bbox=dict(
                        boxstyle="round,pad=0.2",
                        fc="white",
                        ec="k",
                        lw=1,
                        alpha=0.8
                    ),
                    fontsize=text_fontsize
                )

        ax.legend(
            loc="center left",
            bbox_to_anchor=(right, 0.5),
            bbox_transform=fig.transFigure,
            fontsize=legend_fontsize
        )

    # Remove scales and labels on the coordinate axis
    ax.set_xticks([])
    ax.set_yticks([])

    # Remove the bounding box of the coordinate axis
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)

    plot_end(fig, title, x_name, y_name, output, show, close)


def scatter_3d(
    df: DataFrame,
    x: str,
    y: str,
    z: str,
    hue: str = None,
    x_name: str = None,
    y_name: str = None,
    z_name: str = None,
    title: str = None,
    width: float = 7,
    height: float = 7,
    elev: float = 30,
    azim: float = -60,
    is_add_legend: bool = True,
    cmap: Union[str, ListedColormap] = 'tab20',
    font_size: int = 14,
    edge_color: str = None,
    size: Union[float, collection] = 0.1,
    legend_name: str = None,
    is_add_max_label: bool = False,
    text_left_offset: float = 0.5,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
):
    if output is None and not show:
        ul.log(__name__).error(f"At least one of the `output` and `show` parameters is required")
        raise ValueError(f"At least one of the `output` and `show` parameters is required")

    fig = plt.figure(figsize=(width, height))
    ax = fig.add_subplot(projection='3d')

    if hue is not None:
        hue_cat = pd.Categorical(df[hue])

    scatter = ax.scatter(
        df[x],
        df[y],
        df[z],
        c=hue_cat.codes if hue is not None else None,
        cmap=cmap,
        s=size,
        edgecolors=edge_color,
        **kwargs
    )

    # angle of view
    ax.view_init(elev=elev, azim=azim)

    if x_name is not None:
        ax.set_xlabel(x_name, fontsize=font_size)

    if y_name is not None:
        ax.set_ylabel(y_name, fontsize=font_size)

    if z_name is not None:
        ax.set_zlabel(z_name, fontsize=font_size)

    if title is not None:
        ax.set_title(title, fontsize=font_size)

    if is_add_legend and hue is not None:
        unique_types = hue_cat.categories
        legend_elements = [
            plt.Line2D(
                [0], [0], marker='o', color='w', label=type_,
                markerfacecolor=scatter.cmap(scatter.norm(i))
            )
            for i, type_ in enumerate(unique_types)
        ]

        ax.legend(handles=legend_elements, title=legend_name, loc='upper left')

    if is_add_max_label:

        max_idx = df[z].idxmax()
        max_x = df.loc[max_idx, x]
        max_y = df.loc[max_idx, y]
        max_value = df.loc[max_idx, z]

        # 在最大值点的位置添加文本标签
        ax.text(
            max_x - text_left_offset,
            max_y,
            max_value,
            f'({max_x}, {max_y}): {max_value:.3f}',
            fontsize=font_size - 2,
            color='red',
            ha='left'
        )

    plot_end(fig, None, None, None, output, show, close)


def scatter_atac(
    adata: AnnData,
    columns: Tuple[str, str] = ("UMAP1", "UMAP2"),
    clusters: str = "clusters",
    hue_order: list = None,
    width: float = 2,
    height: float = 2,
    x_name: str = None,
    y_name: str = None,
    start_color_index: int = 0,
    color_step_size: int = 0,
    type_colors: collection = None,
    edge_color: str = None,
    size: float = 1.0,
    text_fontsize: float = 7,
    legend_fontsize: float = 7,
    is_text: bool = False,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:
    # DataFrame
    df: DataFrame = adata.obs.copy()
    df[clusters] = df[clusters].astype(str)
    # scatter
    scatter_base(
        df,
        x=columns[0],
        y=columns[1],
        hue=clusters,
        width=width,
        height=height,
        size=size,
        x_name=x_name,
        y_name=y_name,
        hue_order=hue_order,
        start_color_index=start_color_index,
        color_step_size=color_step_size,
        type_colors=type_colors,
        edge_color=edge_color,
        is_text=is_text,
        text_fontsize=text_fontsize,
        legend_fontsize=legend_fontsize,
        output=output,
        show=show,
        close=close,
        right=0.75,
        **kwargs
    )


def scatter_trait(
    trait_adata: AnnData,
    title: str = None,
    bar_label: str = None,
    trait_name: str = "All",
    layers: Union[None, collection] = None,
    columns: Tuple[str, str] = ("UMAP1", "UMAP2"),
    cmap: str = "viridis",
    width: float = 2,
    height: float = 2,
    right: float = 0.9,
    x_name: str = None,
    y_name: str = None,
    number: bool = True,
    edge_color: str = None,
    size: Union[float, collection] = 1.0,
    text_fontsize: float = 7,
    legend_fontsize: float = 7,
    start_color_index: int = 0,
    color_step_size: int = 0,
    type_colors: collection = None,
    is_text: bool = False,
    legend: dict = None,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:
    data: AnnData = trait_adata.copy()

    # judge layers
    trait_adata_layers = list(data.layers)

    if layers is not None and len(layers) != 0:
        for layer in layers:
            if layer not in trait_adata_layers:
                ul.log(__name__).error("The `layers` parameter needs to include in `trait_adata.layers`")
                raise ValueError("The `layers` parameter needs to include in `trait_adata.layers`")

    def trait_plot(trait_: str, atac_cell_df_: DataFrame, layer_: str = None, new_data_: AnnData = None) -> None:
        """
        show plot
        :param trait_: trait name
        :param atac_cell_df_:
        :param layer_: layer
        :param new_data_:
        :return: None
        """
        ul.log(__name__).info(f"Plotting scatter {trait_}")
        # get gene score
        trait_score = new_data_[:, trait_].to_df()
        trait_score = trait_score.rename_axis("__barcode__")
        trait_score.reset_index(inplace=True)
        atac_cell_df_ = atac_cell_df_.rename_axis("__barcode__")
        atac_cell_df_.reset_index(inplace=True)
        # trait_score.rename_axis("index")
        df = atac_cell_df_.merge(trait_score, on="__barcode__", how="left")
        # Sort gene scores from small to large
        df.sort_values([trait_], inplace=True)
        scatter_base(
            df,
            x=columns[0],
            y=columns[1],
            hue=trait_,
            title=f"{title} {trait_}" if title is not None else title,
            bar_label=bar_label,
            legend=legend,
            cmap=cmap,
            width=width,
            height=height,
            right=right,
            number=number,
            size=size,
            x_name=x_name,
            y_name=y_name,
            type_colors=type_colors,
            text_fontsize=text_fontsize,
            legend_fontsize=legend_fontsize,
            start_color_index=start_color_index,
            color_step_size=color_step_size,
            edge_color=edge_color,
            is_text=is_text,
            output=os.path.join(
                output, f"cell_{trait_}_score_{layer_}.pdf" if layer_ is not None else f"cell_{trait_}_score.pdf"
            ) if output is not None else None,
            show=show,
            close=close,
            **kwargs
        )

    def handle_plot(layer_: str = None):
        # DataFrame
        atac_cell_df: DataFrame = data.obs.copy()
        atac_cell_df.rename_axis("index", inplace=True)
        trait_list: list = list(data.var_names)

        # judge trait
        if trait_name != "All" and trait_name not in trait_list:
            ul.log(__name__).error(
                f"The {trait_name} trait/disease is not in the trait/disease list (trait_adata.var_names)")
            raise ValueError(f"The {trait_name} trait/disease is not in the trait/disease list (trait_adata.var_names)")

        new_data: AnnData = AnnData(data.layers[layer], var=data.var, obs=data.obs) if layer_ is not None else data

        # plot
        if trait_name == "All":
            for trait in trait_list:
                trait_plot(trait, atac_cell_df, layer_, new_data)
        else:
            trait_plot(trait_name, atac_cell_df, layer_, new_data)

    if layers is None or len(layers) == 0:
        handle_plot()
    else:
        for layer in layers:
            ul.log(__name__).info(f"Start {layer}")
            handle_plot(layer)


def volcano_base(
    df: DataFrame,
    x: str = "Log2(Fold change)",
    y: str = "-Log10(P value)",
    hue: str = "type",
    size: int = 3,
    palette: Optional[list] = None,
    width: float = 2,
    height: float = 2,
    bottom: float = 0,
    y_min: float = 0,
    axh_value: float = -np.log10(1e-3),
    axv_left_value: float = -1,
    axv_right_value: float = 1,
    title: str = None,
    x_name: Optional[str] = None,
    y_name: Optional[str] = None,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
):
    fig, ax = plot_start(width, height, bottom, output, show)

    if palette is None:
        palette = ["#01c5c4", "#686d76", "#ff414d"]

    sns.set_theme(style="ticks")
    sns.set_palette(sns.color_palette(palette))
    sns.scatterplot(data=df, x=x, y=y, hue=hue, s=size, ax=ax, **kwargs)
    ax.set_ylim(y_min, max(df[y]) * 1.1)

    plt.axhline(axh_value, color='grey', linestyle='--')
    plt.axvline(axv_left_value, color='grey', linestyle='--')
    plt.axvline(axv_right_value, color='grey', linestyle='--')

    plot_end(fig, title, x_name, y_name, output, show, close)


def manhattan_causal_variant(
    df: DataFrame,
    y: str = "pp",
    chr_name: str = "chr",
    label: str = "rsId",
    size: int = 30,
    labels: Optional[list] = None,
    colors: Optional[list] = None,
    width: float = 8,
    height: float = 2,
    bottom: float = 0,
    title: str = None,
    is_sort: bool = True,
    line_width: float = 0.5,
    y_round: int = 3,
    x_name: Optional[str] = "Chromosome",
    y_name: Optional[str] = "pp",
    y_limit: Tuple[float, float] = (0, 1),
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
):
    df[chr_name] = df[chr_name].astype(chrtype)

    if is_sort:
        df = df.sort_values(chr_name)

    df['ind'] = range(len(df))
    df_grouped = df.groupby(chr_name)

    if colors is None:
        colors = type_20_colors.copy()
        colors.extend(type_set_colors)

    fig, ax = plot_start(width, height, bottom, output, show)

    x_labels = []
    x_labels_pos = []
    # Track the last index to draw lines between chromosomes
    last_ind = 0

    chr_unique = df[chr_name].unique()

    for num, (name, group) in enumerate(df_grouped):

        if name not in chr_unique:
            continue

        group.plot(kind='scatter', x='ind', y=y, color=colors[num], s=size, ax=ax)
        x_labels.append(name)
        x_labels_pos.append((group['ind'].iloc[-1] - (group['ind'].iloc[-1] - group['ind'].iloc[0]) / 2))

        # Draw a vertical line between chromosomes
        if num > 0:
            # Skip the first chromosome
            ax.axvline(x=last_ind + 0.5, color='gray', linestyle='--', linewidth=line_width, **kwargs)

        # Label specific mutations
        if labels is not None:
            for index, row in group.iterrows():
                if row[label] in labels:
                    ax.text(row['ind'], row[y], row[label], ha='left', va='bottom')
                    ax.text(row['ind'], row[y], f"{y}={round(row[y], y_round)}", ha='left', va='top')

        last_ind = group['ind'].iloc[-1]

    # add grid
    ax.grid(axis="y", linestyle="--", linewidth=line_width, color="gray")
    ax.set_xticks(x_labels_pos)
    ax.set_xticklabels(x_labels)

    # Hide the borders above and to the right
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)

    ax.set_xlim([0, len(df)])
    ax.set_ylim(y_limit)

    plot_end(fig, title, x_name, y_name, output, show, close)


def pseudo_time_score(
    df: DataFrame,
    x: str,
    y: str,
    x_name: str = None,
    y_name: str = None,
    title: str = None,
    width: float = 2,
    height: float = 1.2,
    bottom: float = 0,
    alpha: float = 0.65,
    line_width: float = 1.5,
    step_length: int = 5,
    polyorder: int = 1,
    size: Union[float, collection] = 1.0,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
):
    from scipy.signal import savgol_filter

    fig, ax = plot_start(width, height, bottom, output, show)

    pseudo_times = df[x].values
    scores = df[y].values

    x_len = len(pseudo_times)

    colors = plt.cm.viridis(np.linspace(0, 1, x_len))

    ax.scatter(
        pseudo_times,
        scores,
        c=colors,
        alpha=alpha,
        s=size,
        **kwargs
    )

    smoothed_scores = savgol_filter(scores, window_length=int(x_len / step_length), polyorder=polyorder)

    ax.plot(pseudo_times, smoothed_scores, color='black', linewidth=line_width)

    plt.tight_layout()

    plot_end(fig, title, x_name, y_name, output, show, close)



# -*- coding: UTF-8 -*-

from typing import Optional, Union, Tuple, Any

import pandas as pd
from anndata import AnnData
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from PyComplexHeatmap import HeatmapAnnotation, anno_simple, ClusterMapPlotter, anno_label, anno_barplot
from matplotlib.colors import ListedColormap
from pandas import DataFrame
import seaborn as sns

from .. import util as ul
from ..util import path, type_20_colors, type_50_colors, plot_end, plot_start

__name__: str = "plot_heat_map"


def heatmap_annotation(
    adata: AnnData,
    layer: Optional[str] = None,
    width: float = 4,
    height: float = 4,
    title: Optional[str] = None,
    label: str = "value",
    row_name: Optional[str] = None,
    col_name: Optional[str] = None,
    row_names: Optional[str] = None,
    col_names: Optional[str] = None,
    row_anno_label: bool = False,
    col_anno_label: bool = False,
    row_anno_text: bool = False,
    col_anno_text: bool = False,
    row_legend: bool = False,
    col_legend: bool = False,
    row_show_names: bool = False,
    col_show_names: bool = False,
    row_cluster: bool = False,
    col_cluster: bool = False,
    cluster_method: str = "average",
    cluster_metric: str = "correlation",
    row_names_side: str = "left",
    col_names_side: str = "bottom",
    bottom: float = 0.01,
    label_size: float = 9,
    fontsize: float = 9,
    level_bar_height: float = None,
    anno_specific_labels: list = None,
    x_label_rotation: float = 245,
    y_label_rotation: float = 0,
    row_color_start_index: int = 0,
    col_color_start_index: int = 10,
    row_split: Union[int, pd.Series] = None,
    col_split: Union[int, pd.Series] = None,
    row_split_order: Union[list, str] = None,
    col_split_order: Union[list, str] = None,
    row_split_gap: float = 0.5,
    col_split_gap: float = 0.2,
    frac: float = 0.2,
    relpos: Tuple = (0, 1),
    anno_label_height: Optional[float] = None,
    selected_anno_label_height: float = 2.5,
    category_height: Optional[float] = 2.5,
    x_name: Optional[str] = None,
    y_name: Optional[str] = None,
    row_score_name: str = "association_score",
    cmap: str = "Oranges",
    is_sort: bool = True,
    show: bool = True,
    close: bool = False,
    output: path = None,
    **kwargs
) -> None:
    """
    Generate a heatmap with row and column annotations.
    :param row_score_name:
    :param category_height:
    :param selected_anno_label_height:
    :param anno_label_height:
    :param frac:
    :param relpos:
    :param level_bar_height:
    :param anno_specific_labels:
    :param adata: input data;
    :param layer: Layer for processing data;
    :param width: The width of the file or image;
    :param height: The height of the file or image;
    :param title: Title of the image;
    :param label: Explanation (title) of the heatmap display value icon in the figure;
    :param row_name: Information on row annotations in the figure;
    :param col_name: Information on column annotations in the figure;
    :param row_names: The information of the row in the picture;
    :param col_names: The information of the column in the figure;
    :param row_anno_label: Whether to display the label of row comments;
    :param col_anno_label: Whether to display the label of column comments;
    :param row_anno_text: Whether to display row comment information in row comments;
    :param col_anno_text: Whether to display column comment information in column comments;
    :param row_legend: Whether to display the category description of the row in the row comments;
    :param col_legend: Whether to display the category description of the column in the column comments;
    :param row_show_names: If set to `true`, display the name of each element in the row;
    :param col_show_names: If set to `true`, display the name of each element in the column;
    :param row_cluster: Set to `true`, perform row clustering;
    :param col_cluster: Set to `true`, perform column clustering;
    :param cluster_method: If `row_cluster` or `col_cluster` is true, the clustering method will be used;
    :param cluster_metric: If `row_cluster` or `col_cluster` is true, the clustering method will be used;
    :param row_names_side: Direction of row names display. Effective when `row_names_side` is `true`;
    :param col_names_side: Direction of column names display. Effective when `col_show_names` is `true`;
    :param bottom: The gap at the bottom in the picture.
    :param label_size: The size of the font for row or column names. Effective when `row_names_side` or `col_show_names` is `true`;
    :param fontsize: The size of the row or column title. Effective when `x_name` or `y_name` not is `None`;
    :param x_label_rotation: The degree of rotation for row names. Effective when `row_names_side` is `true`;
    :param y_label_rotation: The degree of rotation for column names. Effective when `col_show_names` is `true`;
    :param row_color_start_index: Row annotation specifies the starting index of different colors;
    :param col_color_start_index: Column annotation specifies the starting index of different colors;
    :param row_split_order: list or str
        a list to specify the order of row_split, could also be 'cluster_between_groups', if cluster_between_groups was specified,
        hierarchical clustering will be performed on the mean values for each groups and pass the clustered order to row_split_order.
        For example,see https://dingwb.github.io/PyComplexHeatmap/build/html/notebooks/advanced_usage.html#Cluster-between-groups
    :param col_split_order: list or str
        a list to specify the order of col_split, could also be 'cluster_between_groups', if cluster_between_groups was specified,
        hierarchical clustering will be performed on the mean values for each groups and pass the clustered order to row_split_order.
    :param row_split_gap: default are 0.5 and 0.2 mm for row and col.
    :param col_split_gap: default are 0.5 and 0.2 mm for row and col.
    :param row_split: int or pd.Series or pd.DataFrame
        number of cluster for hierarchical clustering or pd.Series or pd.DataFrame, used to split rows or rows into subplots.
    :param col_split: int or pd.Series or pd.DataFrame
        int or pd.Series or pd.DataFrame, used to split rows or columns into subplots.
    :param x_name: Title of row name;
    :param y_name: Title of column name;
    :param cmap: Display color themes for heat maps;
    :param is_sort: If set to true, when displaying the heatmap, the row and column names are sorted and displayed;
    :param show: If true, display the image;
    :param close: If true, close the image;
    :param output: Output file for image saving;
    :return: Display of image or saved file.
    """

    ul.log(__name__).info("Start plotting the heatmap")
    fig, ax = plot_start(width, height, bottom, output, show)

    data = adata.copy()

    # judge layers
    if layer is not None:

        if layer not in list(data.layers):
            ul.log(__name__).error("The value of the `layer` parameter must be one of the keys in `adata.layers`.")
            raise ValueError(f"The `{layer}` parameter needs to include in `adata.layers`")

        data.X = data.layers[layer]

    if is_sort:
        data = data[data.obs.sort_values(row_name).index, data.var.sort_values(col_name).index]

    # DataFrame
    df: DataFrame = data.to_df()

    row_anno: DataFrame = data.obs.copy()

    col_anno: DataFrame = data.var.copy()

    if row_names is not None:
        df.index = data.obs[row_names].astype(str)
        row_anno.index = data.obs[row_names].astype(str)

    if col_names is not None:
        df.columns = data.var[col_names].astype(str)
        col_anno.index = data.var[col_names].astype(str)

    if row_name is not None:
        row_colors = type_20_colors[row_color_start_index:] if len(
            list(set(row_anno[row_name]))) + row_color_start_index <= 20 else type_50_colors[row_color_start_index:]
    else:
        row_colors = "cmap50"

    if col_name is not None:
        col_colors = type_20_colors[col_color_start_index:] if len(
            list(set(col_anno[col_name]))) + col_color_start_index <= 20 else type_50_colors[col_color_start_index:]
    else:
        col_colors = "cmap50"

    df_rows = None
    if anno_specific_labels is not None:
        df_rows = df.apply(lambda x: x.name if x.name in anno_specific_labels else None, axis=1)
        df_rows.name = "Selected"

    # noinspection PyTypeChecker
    row_ha = HeatmapAnnotation(
        label=anno_label(
            row_anno[row_name], cmap=ListedColormap(row_colors), merge=True, height=anno_label_height
        ) if row_anno_label and row_name else None,
        RowCategory=anno_simple(
            row_anno[row_name],
            cmap=ListedColormap(row_colors),
            height=category_height,
            legend=row_legend,
            add_text=row_anno_text,
            text_kws=dict(color="black", rotation=0, fontsize=label_size),
        ) if row_name is not None else None,
        axis=0,
        verbose=0,
        legend_gap=5,
        hgap=0.5,
        label_kws=dict(color="black", rotation=90, horizontalalignment="left")
    )

    # noinspection PyTypeChecker
    row_ha_right = HeatmapAnnotation(
        AssociationScore=anno_barplot(row_anno[[row_score_name]], legend=True, height=level_bar_height,
                                      **dict(edgecolor='none')) if row_score_name in row_anno.columns else None,
        selected=anno_label(df_rows, relpos=relpos, frac=frac, fontsize=label_size,
                            height=selected_anno_label_height) if anno_specific_labels is not None else None,
        axis=0,
        verbose=0,
        legend_gap=5,
        hgap=0.5,
        label_kws=dict(color="black", rotation=90, horizontalalignment="left")
    )

    col_ha_args = {"rotation": 90}
    # noinspection PyTypeChecker
    col_ha = HeatmapAnnotation(
        label=anno_label(
            col_anno[col_name], cmap=ListedColormap(col_colors), merge=True, height=anno_label_height, **col_ha_args
        ) if col_anno_label and col_name else None,
        ColCategory=anno_simple(
            col_anno[col_name],
            cmap=ListedColormap(col_colors),
            height=category_height,
            add_text=col_anno_text,
            legend=col_legend,
            text_kws={'fontsize': label_size}
        ) if col_name is not None else None,
        axis=1,
        verbose=0,
        legend_gap=5,
        hgap=0.5,
        label_side='left',
        label_kws=dict(color="black", rotation=0, horizontalalignment="right")
    )

    """
    It is worth noting here that, `row_cluster_metric="correlation"`, When the default parameter 
    `row_cluster_metric` in method `ClusterMapPlotter` is passed into method `distance.pdist`, 
    that is `metric='correlation'`, and this method derives from this 
    `https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.spatial.distance.pdist.html`,
    it can be inferred that there is a division formula in one step, which may result in the possibility of `NA`.
    
    For example, in this scATAC-seq data, if there are two or more traits without any intersection, 
    the denominator will appear as zero.
    
    Therefore, use the `"median"` value for parameter `row_cluster_method`
    Therefore, use the `"euclidean"` value for parameter `row_cluster_metric`
    """

    ClusterMapPlotter(
        data=df,
        top_annotation=col_ha if col_name is not None else None,
        left_annotation=row_ha if row_name is not None else None,
        right_annotation=row_ha_right if anno_specific_labels is not None or row_score_name in row_anno.columns else None,
        label=label,
        row_cluster_method=cluster_method,
        row_cluster_metric=cluster_metric,
        col_cluster_method=cluster_method,
        col_cluster_metric=cluster_metric,
        show_rownames=row_show_names,
        show_colnames=col_show_names,
        row_names_side=row_names_side,
        col_names_side=col_names_side,
        col_split=col_split,
        row_split=row_split,
        row_split_order=row_split_order,
        col_split_order=col_split_order,
        col_split_gap=col_split_gap,
        row_split_gap=row_split_gap,
        xticklabels_kws=dict(labelrotation=x_label_rotation, labelcolor='black', labelsize=label_size),
        yticklabels_kws=dict(labelrotation=y_label_rotation, labelcolor='black', labelsize=label_size),
        cmap=cmap,
        tree_kws={'row_cmap': 'Dark2'},
        xlabel=x_name,
        ylabel=y_name,
        xlabel_kws=dict(color='black', fontsize=fontsize),
        ylabel_kws=dict(color='black', fontsize=fontsize),
        col_cluster=col_cluster,
        row_cluster=row_cluster,
        col_dendrogram=col_cluster,
        row_dendrogram=row_cluster,
        **kwargs
    )

    plot_end(fig, title, x_name, y_name, output, show, close)


def heatmap(
    adata: AnnData,
    layer: str = None,
    title: Optional[str] = None,
    width: float = 4,
    height: float = 4,
    bottom: float = 0,
    annot: bool = False,
    square: bool = True,
    is_cluster: bool = False,
    cmap: str = "Oranges",
    line_widths: float = 1,
    fmt: str = ".2f",
    rotation: float = 65,
    x_name: str = None,
    y_name: str = None,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:
    fig, ax = plot_start(width, height, bottom, output, show)

    data = adata.copy()

    # judge layers
    if layer is not None:

        if layer not in list(data.layers):
            ul.log(__name__).error("The value of the `layer` parameter must be one of the keys in `adata.layers`.")
            raise ValueError("The value of the `layer` parameter must be one of the keys in `adata.layers`.")

        data.X = data.layers[layer]

    # DataFrame
    ul.log(__name__).info(f"to DataFrame")
    df: DataFrame = data.to_df()
    # seaborn
    heat_map: Axes = sns.clustermap(data=df, square=square, annot=annot, cmap=cmap, fmt=fmt, **kwargs) \
        if is_cluster else \
        sns.heatmap(data=df, square=square, annot=annot, cmap=cmap, linewidths=line_widths, fmt=fmt, **kwargs)

    if not is_cluster:
        plt.setp(heat_map.get_xticklabels(), rotation=rotation, ha="right", rotation_mode="anchor")
    else:
        # noinspection PyUnresolvedReferences
        plt.setp(heat_map.ax_heatmap.get_xticklabels(), rotation=rotation)

    plot_end(fig, title, x_name, y_name, output, show, close)



# -*- coding: UTF-8 -*-

import os
from typing import Tuple, Union, Literal, Any

from pandas import DataFrame
import seaborn as sns

from .. import util as ul
from ..util import path, plot_end, plot_start

__name__: str = "plot_violin"

_Kind = Literal["strip", "swarm", "box", "violin", "boxen", "point", "bar", "count"]


def violin_base(
    df: DataFrame,
    value: str = "value",
    x_name: str = None,
    y_name: str = "value",
    kind: _Kind = "violin",
    clusters: str = "clusters",
    palette: Union[Tuple, list] = None,
    hue: str = None,
    width: float = 2,
    height: float = 2,
    bottom: float = 0.3,
    rotation: float = 65,
    line_width: float = 0.5,
    title: str = None,
    split: bool = False,
    is_sort: bool = True,
    order_names: list = None,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:
    # judge
    df_columns = list(df.columns)

    if value not in df_columns:
        ul.log(__name__).error(
            f"The `value` ({value}) parameter must be in the `df` parameter data column name ({df_columns})")
        raise ValueError(
            f"The `value` ({value}) parameter must be in the `df` parameter data column name ({df_columns})")

    if hue is not None and hue not in df_columns:
        ul.log(__name__).error(
            f"The `hue` ({hue}) parameter must be in the `df` parameter data column name ({df_columns})"
        )
        raise ValueError(f"The `hue` ({hue}) parameter must be in the `df` parameter data column name ({df_columns})")

    fig, ax = plot_start(width, height, bottom, output, show)

    group_columns = [clusters]

    new_df: DataFrame = df.groupby(group_columns, as_index=False)[value].median()

    if "color" in df_columns:
        new_df_color: DataFrame = df.groupby(group_columns, as_index=False)["color"].first()
        new_df = new_df.merge(new_df_color, how="left", on=clusters)

    colors: list = []

    # sort
    if is_sort:
        new_df.sort_values([value], ascending=False, inplace=True)
        y_names: Union[list, None] = list(new_df[clusters])

        if "color" in df_columns:
            colors = list(new_df["color"])

    else:
        new_df.index = new_df[clusters]

        if order_names is not None:
            y_names: list = order_names

            if "color" in df_columns:

                for i in order_names:

                    for j, c in zip(new_df[clusters], new_df["color"]):

                        if i == j:
                            colors.append(c)
                            break

        else:
            y_names = list(new_df[clusters])

            if "color" in df_columns:
                colors = list(new_df["color"])

    # scatter
    g = sns.catplot(
        data=df,
        x=clusters,
        y=value,
        kind=kind,
        hue=hue,
        order=y_names,
        split=split,
        linewidth=line_width,
        palette=palette if palette is not None else (colors if "color" in df_columns else None),
        **kwargs
    )

    # set coordinate
    for _ax_ in g.axes.flat:
        _ax_.spines['top'].set_linewidth(line_width)
        _ax_.spines['right'].set_linewidth(line_width)
        _ax_.spines['bottom'].set_linewidth(line_width)
        _ax_.spines['left'].set_linewidth(line_width)
        # Set the rotation angle of the x-axis labels
        _ax_.tick_params(axis='x', rotation=rotation)

    plot_end(fig, title, x_name, y_name, output, show, close)


def violin_trait(
    trait_df: DataFrame,
    trait_name: Union[str, list] = "All",
    trait_column_name: str = "id",
    value: str = "value",
    clusters: str = "clusters",
    kind: _Kind = "violin",
    x_name: str = None,
    y_name: str = "value",
    palette: Tuple = None,
    width: float = 2,
    height: float = 2,
    rotation: float = 65,
    line_width: float = 0.1,
    bottom: float = 0.3,
    split: bool = False,
    is_sort: bool = True,
    order_names: list = None,
    title: str = None,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:

    data: DataFrame = trait_df.copy()

    def trait_plot(_trait_: Union[str, list], _cell_df_: DataFrame) -> None:
        """
        show plot
        :param _trait_: trait name
        :param _cell_df_:
        :return: None
        """
        ul.log(__name__).info("Plotting box {}".format(_trait_))
        # get gene score
        _filename_: str = _trait_
        trait_score = _cell_df_[_cell_df_[trait_column_name] == _trait_]
        # Sort gene scores from small to large
        violin_base(
            df=trait_score,
            value=value,
            x_name=x_name,
            y_name=y_name,
            width=width,
            palette=palette,
            height=height,
            bottom=bottom,
            split=split,
            is_sort=is_sort,
            rotation=rotation,
            order_names=order_names,
            kind=kind,
            hue=trait_column_name,
            line_width=line_width,
            clusters=clusters,
            title=f"{title} {_filename_}" if title is not None else title,
            output=os.path.join(output, f"cell_{_filename_}_score_cat_{kind}.pdf") if output is not None else None,
            show=show,
            close=close,
            **kwargs
        )

    # noinspection DuplicatedCode
    trait_list = list(set(data[trait_column_name]))
    # judge trait
    if trait_name != "All":
        if isinstance(trait_name, str):
            if trait_name not in trait_list:
                ul.log(__name__).error(f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}.")
                raise ValueError(f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}.")
        else:
            for tn in trait_name:
                if tn not in trait_list:
                    ul.log(__name__).error(f"The {tn} trait/disease is not in the trait/disease list {trait_list}.")
                    raise ValueError(f"The {tn} trait/disease is not in the trait/disease list {trait_list}.")

    # plot
    if trait_name == "All":
        for trait in trait_list:
            trait_plot(trait, trait_df)
    else:
        trait_plot(trait_name, trait_df)



# -*- coding: UTF-8 -*-

import math
import random
from typing import Optional, Literal, Union

import matplotlib
import networkx as nx
import numpy as np
from anndata import AnnData
from matplotlib import pyplot as plt, gridspec

from .. import util as ul
from ..util import (
    matrix_data,
    collection,
    path,
    list_index,
    to_dense,
    type_20_colors,
    type_50_colors,
    check_adata_get,
    plot_end, plot_start
)

__name__: str = "plot_graph"

_LayoutType = Optional[Literal['spring', 'kamada_kawai', 'circular', 'shell', 'circular_type1', 'circular_type2', 'square_type1', 'square_type2']]


def graph(
    data: matrix_data,
    labels: collection = None,
    node_size: int = 50,
    name: str = None,
    x_name: str = None,
    y_name: str = None,
    title: str = None,
    width: float = 2,
    height: float = 2,
    bottom: float = 0,
    is_font: bool = False,
    output: path = None,
    show: bool = True,
    close: bool = False
) -> None:
    if output is None and not show:
        ul.log(__name__).error(f"At least one of the `output` and `show` parameters is required")
        raise ValueError(f"At least one of the `output` and `show` parameters is required")

    plt.figure(figsize=(width, height), dpi=300)

    fig, ax = plot_start(width, height, bottom, output, show)

    # Determine whether it is a square array
    if data.shape[0] != data.shape[1]:
        ul.log(__name__).error("The input data must be a square matrix.")
        raise ValueError("The input data must be a square matrix.")

    # set labels
    labels_dict = {}

    if labels is not None:

        if data.shape[0] != np.asarray(labels).size:
            ul.log(__name__).error(
                f"The number of input data nodes {data.shape[0]} and the number of "
                f"labels {np.asarray(labels).size} must be consistent"
            )
            raise ValueError(
                f"The number of input data nodes {data.shape[0]} and the number of "
                f"labels {np.asarray(labels).size} must be consistent"
            )

        labels_dict: dict = dict(zip(range(len(labels)), labels))

    rows, cols = np.where(data == 1)
    edges = zip(rows.tolist(), cols.tolist())
    gr = nx.Graph(name=name)
    gr.add_edges_from(edges)
    pos = nx.spring_layout(gr, k=0.15, iterations=35, seed=2023)

    options: dict = {
        "node_color": "black",
        "node_size": node_size,
        "linewidths": 0,
        "width": 0.1
    }

    if is_font:
        if labels is not None:
            nx.draw(gr, pos=pos, labels=labels_dict, **options)
        else:
            nx.draw(gr, pos=pos, **options)
    else:
        nx.draw(gr, pos=pos, labels={}, **options)

    plot_end(fig, title, x_name, y_name, output, show, close)


def communities_graph(
    adata: AnnData,
    labels: collection,
    layer: str = None,
    clusters: str = "clusters",
    x_name: str = None,
    y_name: str = None,
    title: str = None,
    width: float = 2,
    height: float = 2,
    bottom: float = 0,
    node_size: float = 2.0,
    line_widths: float = 0.001,
    start_color_index: int = 0,
    color_step_size: int = 0,
    output: path = None,
    show: bool = True,
    close: bool = False
):
    if output is None and not show:
        ul.log(__name__).error(f"At least one of the `output` and `show` parameters is required")
        raise ValueError(f"At least one of the `output` and `show` parameters is required")

    ul.log(__name__).info("Start cell-cell network diagram")

    new_data = check_adata_get(adata=adata, layer=layer)

    # adjust matrix
    adj_matrix = to_dense(new_data.X)
    communities, node_labels = list_index(labels)

    df = new_data.obs.copy()

    __hue_order__ = list(np.sort(list(set(df[clusters]))))

    type_colors = type_20_colors if len(__hue_order__) <= 20 else type_50_colors

    fig, ax = plot_start(width, height, bottom, output, show)

    ul.log(__name__).info("Get position")
    color_index = 0
    g = nx.from_numpy_array(adj_matrix)
    partition: list = [0 for _ in range(g.number_of_nodes())]

    for c_i, nodes in enumerate(communities):

        for i in nodes:
            partition[i] = type_colors[start_color_index + color_index * color_step_size + c_i]

        color_index += 1

    pos = nx.spring_layout(g)

    pos1 = [p[0] for p in pos.values()]
    pos2 = [p[1] for p in pos.values()]
    new_data.obs["pos1"] = pos1
    new_data.obs["pos2"] = pos2

    plt.axis("off")

    nx.draw_networkx_nodes(
        g,
        pos=pos,
        node_size=node_size,
        node_color=partition,
        linewidths=line_widths
    )
    # nodes.set_edgecolor("b")
    nx.draw_networkx_edges(
        g,
        pos=pos,
        node_size=node_size,
        edge_color=(0, 0, 0, 0.3),
        width=line_widths
    )

    plot_end(fig, title, x_name, y_name, output, show, close)


def network_two_types(
    data_pairs: list,
    type1_scores: dict,
    type2_scores: dict,
    type1_node_size: Optional[Union[dict, list, float]] = 50,
    type2_node_size: Optional[Union[dict, list, float]] = 50,
    label_nodes: Optional[list] = None,
    width: float = 4,
    height: float = 3,
    k: Optional[float] = None,
    iterations: int = 50,
    scale: float = 1,
    radius: float = 0.35,
    type1_node_shape: str = 'o',
    type2_node_shape: str = 's',
    type1_bar_label: str = 'Score',
    type2_bar_label: str = 'Score',
    type1_cmap_str: str = "winter",
    type2_cmap_str: str = "YlOrRd",
    node_alpha: float = 0.8,
    edge_alpha: float = 0.8,
    is_fluctuate: bool = True,
    layout_type: str = 'spring',
    output: path = None,
    show: bool = True,
    close: bool = False
):
    if output is None and not show:
        ul.log(__name__).error(f"At least one of the `output` and `show` parameters is required")
        raise ValueError(f"At least one of the `output` and `show` parameters is required")

    # Create a node list of genes and variations
    type1_nodes = []
    type2_nodes = []

    type1_keys = type1_scores.keys()
    type2_keys = type2_scores.keys()

    for type1_node, type2_node in data_pairs:

        if type1_node not in type1_keys:
            raise Exception(f"Node `{type1_node}` not in `type1_scores`.")

        if type2_node not in type2_keys:
            raise Exception(f"Node `{type2_node}` not in `type2_scores`.")

        if type1_node not in type1_nodes:
            type1_nodes.append(type1_node)

        if type2_node not in type2_nodes:
            type2_nodes.append(type2_node)

    # Create graphics and grids
    fig = plt.figure(figsize=(width, height))
    gs = gridspec.GridSpec(
        2, 2,
        width_ratios=[7, 1],
        height_ratios=[4, 4],
        hspace=0.1,
        wspace=0.1
    )

    # Spindle is used to draw network diagrams
    ax_network = fig.add_subplot(gs[0:2, 0:1])

    # Create a directed graph
    G = nx.Graph()
    G.add_edges_from(data_pairs)

    # draw graphics
    if layout_type == 'spring':
        pos = nx.spring_layout(G, k=k, iterations=iterations)
    elif layout_type == 'kamada_kawai':
        pos = nx.kamada_kawai_layout(G, scale=scale)
    elif layout_type == 'circular':
        pos = nx.circular_layout(G, scale=scale)
    elif layout_type == 'shell':
        shells = [type1_nodes, type2_nodes]
        pos = nx.shell_layout(G, shells, scale=scale)
    else:
        if layout_type == 'circular_type1':
            is_type1 = True
            type1_positions = nx.circular_layout(G.subgraph(type1_nodes), scale=scale)
            pos = {**type1_positions}
        elif layout_type == 'circular_type2':
            is_type1 = False
            type2_positions = nx.circular_layout(G.subgraph(type2_nodes), scale=scale)
            pos = {**type2_positions}
        else:
            if layout_type == 'square_type1':
                is_type1 = True
            elif layout_type == 'square_type2':
                is_type1 = False
            else:
                raise ValueError("The `layout_type` parameter must be one of the following string values {'spring','kamada_kawai','circular','shell','circular_type1','circular_type2','square_type1','square_type2'}")

            # Adjust to square coordinates
            pos = {}
            side_length = int(np.sqrt(len(type1_nodes if is_type1 else type2_nodes)))
            for i, key in enumerate(type1_nodes if is_type1 else type2_nodes):
                pos[key] = (i % side_length, i // side_length)

        for i, _node_ in enumerate(type1_nodes if is_type1 else type2_nodes):
            type_other = []
            for k, v in data_pairs:
                if is_type1:
                    if k == _node_:
                        type_other.append(v)
                else:
                    if v == _node_:
                        type_other.append(k)

            angle_step = 360 / len(type_other)
            for j, _type_node_ in enumerate(type_other):
                angle = math.radians(j * angle_step) + (random.random() / 2 if is_fluctuate else 0)
                x = pos[_node_][0] + np.power(-1, int(random.random() * 10) if is_fluctuate else 0) * (radius + (random.random() / 10 if is_fluctuate else 0)) * math.cos(angle)
                y = pos[_node_][1] + np.power(-1, int(random.random() * 10) if is_fluctuate else 0) * (radius + (random.random() / 10 if is_fluctuate else 0)) * math.sin(angle)
                pos[_type_node_] = (x, y)

    type1_cmap = matplotlib.colormaps[type1_cmap_str]
    type2_cmap = matplotlib.colormaps[type2_cmap_str]

    # Ensure that the normalized range of color mapping matches the value range of gene scores
    type1_norm = plt.Normalize(vmin=min(type1_scores.values()), vmax=max(type1_scores.values()))
    type2_norm = plt.Normalize(vmin=min(type2_scores.values()), vmax=max(type2_scores.values()))

    type1_node_colors = [type1_cmap(type1_norm(type1_scores[node])) for node in type1_nodes]
    nx.draw_networkx_nodes(
        G,
        pos,
        nodelist=type1_nodes,
        node_size=[type1_node_size[node] for node in type1_nodes] if isinstance(type1_node_size, dict) else type1_node_size,
        node_color=type1_node_colors,
        edgecolors='#333333',
        node_shape=type1_node_shape,
        alpha=node_alpha,
        ax=ax_network
    )

    type2_node_colors = [type2_cmap(type2_norm(type2_scores[node])) for node in type2_nodes]
    nx.draw_networkx_nodes(
        G,
        pos,
        nodelist=type2_nodes,
        node_size=[type2_node_size[node] for node in type2_nodes] if isinstance(type2_node_size, dict) else type2_node_size,
        node_color=type2_node_colors,
        edgecolors='#333333',
        node_shape=type2_node_shape,
        alpha=node_alpha,
        ax=ax_network
    )

    if label_nodes is not None:
        labels = {node: node for node in label_nodes}

        # Draw node labels
        nx.draw_networkx_labels(
            G,
            pos,
            labels,
            font_size=10,
            ax=ax_network,
            verticalalignment='center',
            horizontalalignment='center'
        )

    # Color and width of edges
    edge_colors = ['gray' for _ in data_pairs]
    edge_widths = [1.0 for _ in data_pairs]

    # Draw edges
    nx.draw_networkx_edges(
        G,
        pos,
        edge_color=edge_colors,
        width=edge_widths,
        alpha=edge_alpha,
        arrows=False,
        ax=ax_network,
        style='solid'
    )

    # Set background color
    ax_network.set_facecolor('white')
    ax_network.spines['top'].set_visible(False)
    ax_network.spines['right'].set_visible(False)

    # Color bar axis
    ax_cbar1 = fig.add_subplot(gs[0:1, 1:2])
    ax_cbar1.set_axis_off()
    # Create a color bar
    sm1 = plt.cm.ScalarMappable(cmap=type1_cmap, norm=type1_norm)
    cbar1 = plt.colorbar(sm1, ax=ax_cbar1, location='right', pad=0.2)
    cbar1.set_label(type1_bar_label)

    ax_cbar2 = fig.add_subplot(gs[1:2, 1:2])
    ax_cbar2.set_axis_off()
    sm2 = plt.cm.ScalarMappable(cmap=type2_cmap, norm=type2_norm)
    cbar2 = plt.colorbar(sm2, ax=ax_cbar2, location='right', pad=0.2)
    cbar2.set_label(type2_bar_label)

    plt.axis('off')

    plot_end(fig, output=output, show=show, close=close)
